unit XmlIni;

interface

{
  Design By Howard Chen.
  EMail: howardchen@pie.com.tw
  MSN: powerhowardchen@hotmail.com

  XmlIni version 5.
}

{
  AES Functions.
  Copyright (c) 1998-2001
  EldoS, Alexander Ionov
}

// Compiler with RSA Encrypt function
{$DEFINE RSA_ENCRYPT_ENABLED}

uses
  Windows, Classes, XMLIntf, Dialogs
  {$IFDEF RSA_ENCRYPT_ENABLED}
  , FGInt
  {$ENDIF}
  ;

type
  TXmlIni = class;
  TXmlIniSection = class;
  TXmlIniKey = class;

  TXmlIniEncodingType = (xietUnknow, xietUTF_8, xietUTF_16,
    xietBIG5, xietGB2312, xietS_JIS, xietEUC_KR, xietIS0_8859_1);

  TXmlIniDataType = (xidtUnknow, xidtText, xidtInteger, xidtFloat, xidtBool, xidtDateTime, xidtDate, xidtTime, xidtBinary);

  TXmlIniEncryptMode = (xiemAuto, xiemPlain, xiemAES
{$IFDEF RSA_ENCRYPT_ENABLED}
  , xiemRSA
{$ENDIF}
  );

  TXmlIniEventMemoryStream = class(TMemoryStream)
  private
    FOwner: TXmlIniKey;
  public
    constructor Create(AOwner: TXmlIniKey);
    function GetSize: Int64; override;
    procedure SetSize(NewSize: Longint); override;
    function Read(var Buffer; Count: Longint): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
  end;

  TXmlIniKey = class(TObject)
  private
    FOwner: TXmlIniSection;
    FName: WideString;
    FText: WideString;
    FStream: TXmlIniEventMemoryStream;
    FDataType: TXmlIniDataType;
    FRaiseException: Boolean;

    procedure BinaryToText;
    procedure TextToBinary;
    procedure SetName(const Value: WideString);
    function GetText: WideString;
    procedure SetText(const Value: WideString);
    function GetAsStream: TStream;
    procedure SetAsStream(const Value: TStream);
    function GetAsInteger: Integer;
    procedure SetAsInteger(const Value: Integer);
    function GetAsFloat: Double;
    procedure SetAsFloat(const Value: Double);
    function GetAsBool: Boolean;
    procedure SetAsBool(const Value: Boolean);
    function GetAsDateTime: TDateTime;
    procedure SetAsDateTime(const Value: TDateTime);
    function GetAsDate: TDateTime;
    procedure SetAsDate(const Value: TDateTime);
    function GetAsTime: TDateTime;
    procedure SetAsTime(const Value: TDateTime);
    function GetDataType: TXmlIniDataType;
    procedure SetDataType(const Value: TXmlIniDataType);
  public
    constructor Create(AName: WideString = ''; AText: WideString = ''; AOwner: TXmlIniSection = nil);
    destructor Destroy; override;

    procedure Clear;

    procedure ValueLoadFromFile(FileName: WideString; Binary: Boolean = True); virtual;
    procedure ValueSaveToFile(FileName: WideString); virtual;

    property Name: WideString read FName write SetName;
    property Text: WideString read GetText write SetText;
    property AsStream: TStream read GetAsStream write SetAsStream;
    property AsInteger: Integer read GetAsInteger write SetAsInteger;
    property AsFloat: Double read GetAsFloat write SetAsFloat;
    property AsBool: Boolean read GetAsBool write SetAsBool;
    property AsDateTime: TDateTime read GetAsDateTime write SetAsDateTime;
    property AsDate: TDateTime read GetAsDate write SetAsDate;
    property AsTime: TDateTime read GetAsTime write SetAsTime;
    property DataType: TXmlIniDataType read GetDataType write SetDataType;
    property RaiseException: Boolean read FRaiseException write FRaiseException;
  end;

  TXmlIniSection = class(TObject)
  private
    FName: WideString;
    FOwner: TXmlIni;
    FList: TList;
    FXmlEncoding: TXmlIniEncodingType;
    FCaseSensitive: Boolean;
    FCompressed: Boolean;
    FRaiseException: Boolean;

    function InternalFind(KeyName: WideString): TXmlIniKey;
    function InternalAdd(KeyName: WideString): TXmlIniKey;
    procedure SetName(const Value: WideString);
    function GetKeyByIndex(Index: Integer): TXmlIniKey;
    function GetKey(Index: WideString): TXmlIniKey;
    function GetItem(Index: WideString): WideString;
    procedure SetItem(Index: WideString; const Value: WideString);
    function GetCount: Integer;
    procedure SetRaiseException(const Value: Boolean);
  public
    constructor Create(AName: WideString = ''; AOwner: TXmlIni = nil);
    destructor Destroy; override;

    function Equals(Dest: TXmlIniSection): Boolean; overload;
    procedure CopyFrom(Source: TXmlIniSection); overload;
    procedure Append(Source: TXmlIniSection); overload;

    function Equals(Dest: TStrings): Boolean; overload;
    procedure CopyFrom(Source: TStrings); overload;
    procedure Append(Source: TStrings); overload;

    function Exists(KeyName: WideString): Boolean;

    procedure Clear;
    procedure Collation;

    procedure LoadFromStream(Stream: TStream; SectionName: WideString = ''); virtual;
    procedure SaveToStream(Stream: TStream; SectionName: WideString = ''); virtual;

    procedure LoadFromString(Value: WideString; SectionName: WideString = '');
    function SaveToString(SectionName: WideString = ''): WideString;

    procedure LoadFromBase64String(Value: string; SectionName: WideString = '');
    function SaveToBase64String(SectionName: WideString = ''): string;

    procedure LoadFromFile(FileName: WideString; SectionName: WideString = '');
    procedure SaveToFile(FileName: WideString; SectionName: WideString = '');

    procedure LoadFromRegistry(RootKey: HKEY; Path: WideString; SectionName: WideString = ''); virtual;
    procedure SaveToRegistry(RootKey: HKEY; Path: WideString; SectionName: WideString = '';
      UseOldDataType: Boolean = False); virtual;

    procedure LoadFromIni(FileName: WideString; SectionName: WideString = '');
    procedure SaveToIni(FileName: WideString; SectionName: WideString = '');

    property XmlEncoding: TXmlIniEncodingType read FXmlEncoding write FXmlEncoding;
    property CaseSensitive: Boolean read FCaseSensitive write FCaseSensitive;
    property Compressed: Boolean read FCompressed write FCompressed;

    property Name: WideString read FName write SetName;
    property KeyByIndex[Index: Integer]: TXmlIniKey read GetKeyByIndex;
    property Key[Index: WideString]: TXmlIniKey read GetKey;
    property Item[Index: WideString]: WideString read GetItem write SetItem; default;
    property Count: Integer read GetCount;
    property RaiseException: Boolean read FRaiseException write SetRaiseException;
  end;

  TXmlIni = class(TComponent)
  private
    FList: TList;
    FXmlEncoding: TXmlIniEncodingType;
    FFileName: WideString;
    FAutoSave: Boolean;
    FCaseSensitive: Boolean;
    FCompressed: Boolean;
    FEncryptMode: TXmlIniEncryptMode;
    FPassword: WideString;
    {$IFDEF RSA_ENCRYPT_ENABLED}
    FPrivateKey: TStringList;
    FPublicKey: TStringList;
    {$ENDIF}
    FRaiseException: Boolean;
    FLockCriSec: TRTLCriticalSection;

    procedure InternalXmlLoadFromStream(Stream: TStream);
    procedure InternalXmlSaveToStream(Stream: TStream);
    procedure InternalLoadFromFile(AFileName: WideString);
    procedure InternalSaveToFile(AFileName: WideString);
    function InternalFind(SectionName: WideString): TXmlIniSection;
    function InternalAdd(SectionName: WideString): TXmlIniSection;
    {$IFDEF RSA_ENCRYPT_ENABLED}
    procedure ExtractKey(KeyStr: string; var Key1, Key2: TFGInt);
    {$ENDIF}
    function GetCount: Integer;
    function GetSectionByIndex(Index: Integer): TXmlIniSection;
    function GetSection(Index: WideString): TXmlIniSection;
    procedure SetFileName(const Value: WideString);
    procedure SetAutoSave(const Value: Boolean);
    procedure SetCaseSensitive(const Value: Boolean);
    procedure SetCompressed(const Value: Boolean);
    procedure SetEncryptMode(const Value: TXmlIniEncryptMode);
    procedure SetPassword(const Value: WideString);
    {$IFDEF RSA_ENCRYPT_ENABLED}
    function GetPrivateKey: TStrings;
    procedure SetPrivateKey(const Value: TStrings);
    function GetPublicKey: TStrings;
    procedure SetPublicKey(const Value: TStrings);
    {$ENDIF}
    procedure SetRaiseException(const Value: Boolean);
  public
    constructor Create(AOwner: TComponent); overload; override;
    constructor Create(AFileName: WideString = ''); reintroduce; overload;
    destructor Destroy; override;

    function Equals(Dest: TXmlIni): Boolean;
    procedure CopyFrom(Source: TXmlIni);
    procedure Append(Source: TXmlIni);

    function Add(SectionName: WideString): TXmlIniSection;
    procedure Delete(SectionName: WideString);
    function Exists(SectionName: WideString; KeyName: WideString = ''): Boolean;

    procedure Clear;
    procedure Collation;

    procedure LoadFromStream(Stream: TStream); virtual;
    procedure SaveToStream(Stream: TStream); virtual;

    procedure LoadFromString(Value: WideString);
    function SaveToString: WideString;

    procedure LoadFromBase64String(Value: string);
    function SaveToBase64String: string;

    procedure LoadFromFile(AFileName: WideString = ''); virtual;
    procedure SaveToFile(AFileName: WideString = ''); virtual;

    procedure LoadFromRegistry(RootKey: HKEY; Path: WideString); virtual;
    procedure SaveToRegistry(RootKey: HKEY; Path: WideString); virtual;

    procedure LoadFromIni(AIniFileName: WideString); virtual;
    procedure SaveToIni(AIniFileName: WideString); virtual;

    procedure LoadFromShareMem(ShareMemID: string = ''); virtual;
    procedure SaveToShareMem(ShareMemID: string = ''); virtual;

    {$IFDEF RSA_ENCRYPT_ENABLED}
    procedure GenerateNewKey;
    {$ENDIF}

    procedure Lock; virtual;
    procedure Unlock; virtual;

    // TIniFile Functions Begin
    function SectionExists(const Section: WideString): Boolean;

    function ReadString(const Section, Ident, Default: WideString): WideString; virtual;
    procedure WriteString(const Section, Ident, Value: WideString); virtual;
    function ReadInteger(const Section, Ident: WideString; Default: Longint): Longint; virtual;
    procedure WriteInteger(const Section, Ident: WideString; Value: Longint); virtual;
    function ReadFloat(const Section, Ident: WideString; Default: Double): Double; virtual;
    procedure WriteFloat(const Section, Ident: WideString; Value: Double); virtual;
    function ReadBool(const Section, Ident: WideString; Default: Boolean): Boolean; virtual;
    procedure WriteBool(const Section, Ident: WideString; Value: Boolean); virtual;
    function ReadDateTime(const Section, Ident: WideString; Default: TDateTime): TDateTime; virtual;
    procedure WriteDateTime(const Section, Ident: WideString; Value: TDateTime); virtual;
    function ReadDate(const Section, Ident: WideString; Default: TDateTime): TDateTime; virtual;
    procedure WriteDate(const Section, Ident: WideString; Value: TDateTime); virtual;
    function ReadTime(const Section, Ident: WideString; Default: TDateTime): TDateTime; virtual;
    procedure WriteTime(const Section, Ident: WideString; Value: TDateTime); virtual;
    function ReadBinaryStream(const Section, Ident: WideString; Value: TStream): Integer; virtual;
    procedure WriteBinaryStream(const Section, Ident: WideString; Value: TStream); virtual;

    procedure ReadSection(const Section: WideString; Strings: TStrings); virtual;
    procedure ReadSections(Strings: TStrings); virtual;
    procedure ReadSectionValues(const Section: WideString; Strings: TStrings); virtual;
    procedure EraseSection(const Section: WideString); virtual;

    procedure DeleteKey(const Section, Ident: WideString); virtual;
    procedure UpdateFile; virtual;
    function ValueExists(const Section, Ident: WideString): Boolean;
    // TIniFile Functions End

    property SectionByIndex[Index: Integer]: TXmlIniSection read GetSectionByIndex;
    property Section[Index: WideString]: TXmlIniSection read GetSection; default;
    property Count: Integer read GetCount;
  published
    property FileName: WideString read FFileName write SetFileName;
    property AutoSave: Boolean read FAutoSave write SetAutoSave default False;

    property XmlEncoding: TXmlIniEncodingType read FXmlEncoding write FXmlEncoding default xietUTF_16;
    property CaseSensitive: Boolean read FCaseSensitive write SetCaseSensitive default False;
    property Compressed: Boolean read FCompressed write SetCompressed default False;
    property EncryptMode: TXmlIniEncryptMode read FEncryptMode write SetEncryptMode default xiemAuto;
    property Password: WideString read FPassword write SetPassword;
    {$IFDEF RSA_ENCRYPT_ENABLED}
    property PrivateKey: TStrings read GetPrivateKey write SetPrivateKey;
    property PublicKey: TStrings read GetPublicKey write SetPublicKey;
    {$ENDIF}

    property RaiseException: Boolean read FRaiseException write SetRaiseException default False;
  end;

function ExStrToDateTime(const Value: WideString): TDateTime;
function ExStrToDate(const Value: WideString): TDateTime;
function ExStrToTime(const Value: WideString): TDateTime;
function ExDateTimeToStr(const Value: TDateTime): WideString;
function ExDateToStr(const Value: TDateTime): WideString;
function ExTimeToStr(const Value: TDateTime): WideString;

function Base64ToStr(Base64: WideString): WideString;
function StrToBase64(Str: WideString): WideString;

const
  XI_XML_EncodingStr: array[TXmlIniEncodingType] of WideString =
  ('Unknow', 'UTF-8', 'UTF-16', 'BIG5', 'GB2312', 'S-JIS', 'EUC-KR', 'IS0-8859-1');

implementation

uses
  xmldom, msxmldom, XMLDoc, XSBuiltIns, ActiveX, Forms, zlib, SysUtils, DateUtils, StrUtils, Registry, IniFiles
  {$IFDEF RSA_ENCRYPT_ENABLED}
  , FGIntPrimeGeneration, FGIntRSA
  {$ENDIF}
  ;

const
  XI_XML_StandAlone = 'yes';
  XI_XML_Version = '1.0';

  XI_Root_NodeName = 'xmlini';
  XI_Section_NodeName = 'section';
  XI_Key_NodeName = 'key';
  XI_Name_Attribute = 'name';

  XI_Base64Header = '?BASE64?';

  XI_True_Text: array[0..14] of WideString =
  ('True', 'T', 'Yes', 'Y', 'Just', 'Right', 'Correct', 'OK', '1', 'On', 'Agree', 'O', '', 'n', 'T');
  XI_False_Text: array[0..17] of WideString =
  ('False', 'F', 'No', 'N', 'Unjust', 'Wrong', 'Incorrect', 'Cancel', '0', 'Off', 'Disagree', '_', '',
    '', '', 'n', 'T', '~');

  XI_LockMutexName = 'XMLINI_MUTEX_LOCK_';
  XI_ShareName = 'XMLINI_SHARED_MEM_';

const
  BufSize = 1024;
{$IFDEF RSA_ENCRYPT_ENABLED}
  FakePasswordLength = 16;
  HeaderRSACrypto = 'PC';
{$ENDIF}
  HeaderAESCrypto = 'EC';
  HeaderCompress = 'ZL';

procedure CompressStream(SourceStream, DestStream: TStream);
var
  CS: TCompressionStream;
  Buf: PChar;
  DataSize, ReadCount: LongInt;
begin
  SourceStream.Position := 0;
  DestStream.Size := 0;

  DataSize := SourceStream.Size;
  DestStream.Write(HeaderCompress, 2);
  DestStream.Write(DataSize, SizeOf(LongInt));

  CS := TCompressionStream.Create(clMax, DestStream);
  try
    Buf := AllocMem(BufSize);
    try
      repeat
        ReadCount := SourceStream.Read(Buf^, BufSize);
        if ReadCount > 0 then
          CS.Write(Buf^, ReadCount);
      until ReadCount = 0;
    finally
      FreeMem(Buf);
    end;
  finally
    CS.Free;
  end;
end;

procedure DecompressStream(SourceStream, DestStream: TStream);
var
  DS: TDecompressionStream;
  Ident: array[0..1] of Char;
  Buf: PChar;
  DataSize, ReadCount: LongInt;
begin
  SourceStream.Position := 0;
  SourceStream.Read(Ident, 2);
  SourceStream.Read(DataSize, SizeOf(LongInt));
  if (Ident = HeaderCompress) and (DataSize > 0) then
  begin
    DestStream.Size := 0;
    DS := TDecompressionStream.Create(SourceStream);
    try
      Buf := AllocMem(BufSize);
      try
        while DataSize > 0 do
        begin
          ReadCount := DataSize;
          if ReadCount > BufSize then
            ReadCount := BufSize;
          DS.Read(Buf^, ReadCount);
          DestStream.Write(Buf^, ReadCount);
          Dec(DataSize, ReadCount);
        end;
      finally
        FreeMem(Buf);
      end;
    finally
      DS.Free;
    end;
  end;
end;

type
  PByte4 = ^TByte4;
  TByte4 = packed record
    b1: Byte;
    b2: Byte;
    b3: Byte;
    b4: Byte;
  end;

  PByte3 = ^TByte3;
  TByte3 = packed record
    b1: Byte;
    b2: Byte;
    b3: Byte;
  end;

const
  BUFFER_SIZE = $3000;
 { Caution: For MimeEncodeStream and all other kinds of multi-buffered
   Mime encodings (i.e. Files etc.), BufferSize must be set to a multiple of 3.
   Even though the implementation of the Mime decoding routines below
   does not require a particular buffer size, they work fastest with sizes of
   multiples of four. The chosen size is a multiple of 3 and of 4 as well.
   The following numbers are, in addition, also divisible by 1024:
   $2400, $3000, $3C00, $4800, $5400, $6000, $6C00. }

  MIME_ENCODE_TABLE: array[0..63] of Byte = (
    065, 066, 067, 068, 069, 070, 071, 072, // 00 - 07
    073, 074, 075, 076, 077, 078, 079, 080, // 08 - 15
    081, 082, 083, 084, 085, 086, 087, 088, // 16 - 23
    089, 090, 097, 098, 099, 100, 101, 102, // 24 - 31
    103, 104, 105, 106, 107, 108, 109, 110, // 32 - 39
    111, 112, 113, 114, 115, 116, 117, 118, // 40 - 47
    119, 120, 121, 122, 048, 049, 050, 051, // 48 - 55
    052, 053, 054, 055, 056, 057, 043, 047); // 56 - 63

  MIME_DECODE_TABLE: array[Byte] of Cardinal = (
    255, 255, 255, 255, 255, 255, 255, 255, //  00 -  07
    255, 255, 255, 255, 255, 255, 255, 255, //  08 -  15
    255, 255, 255, 255, 255, 255, 255, 255, //  16 -  23
    255, 255, 255, 255, 255, 255, 255, 255, //  24 -  31
    255, 255, 255, 255, 255, 255, 255, 255, //  32 -  39
    255, 255, 255, 062, 255, 255, 255, 063, //  40 -  47
    052, 053, 054, 055, 056, 057, 058, 059, //  48 -  55
    060, 061, 255, 255, 255, 255, 255, 255, //  56 -  63
    255, 000, 001, 002, 003, 004, 005, 006, //  64 -  71
    007, 008, 009, 010, 011, 012, 013, 014, //  72 -  79
    015, 016, 017, 018, 019, 020, 021, 022, //  80 -  87
    023, 024, 025, 255, 255, 255, 255, 255, //  88 -  95
    255, 026, 027, 028, 029, 030, 031, 032, //  96 - 103
    033, 034, 035, 036, 037, 038, 039, 040, // 104 - 111
    041, 042, 043, 044, 045, 046, 047, 048, // 112 - 119
    049, 050, 051, 255, 255, 255, 255, 255, // 120 - 127
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255);

function MimeEncodedSize(const i: Integer): Integer;
begin
  Result := (i + 2) div 3 * 4;
end;

procedure MimeEncode(const InputBuffer: Pointer; const InputByteCount: Integer; const OutputBuffer: Pointer);
var
  b: Cardinal;
  InMax3: Integer;
  pIn, PInLimit: ^Byte;
  POut: PByte4;
begin
{$IFDEF RangeChecking}
  if InputBuffer = nil then
    raise EMime.Create(SInputBufferNil);
  if OutputBuffer = nil then
    raise EMime.Create(SOutputBufferNil);
{$ENDIF}
  if InputByteCount <= 0 then
    Exit;

  pIn := InputBuffer;
  InMax3 := InputByteCount div 3 * 3;
  POut := OutputBuffer;

  Integer(PInLimit) := Integer(pIn) + InMax3;

  while pIn <> PInLimit do
  begin
    b := pIn^;
    b := b shl 8;
    Inc(pIn);
    b := b or pIn^;
    b := b shl 8;
    Inc(pIn);
    b := b or pIn^;
    Inc(pIn);

   // Write 4 bytes to OutputBuffer (in reverse order).
    POut.b4 := MIME_ENCODE_TABLE[b and $3F];
    b := b shr 6;
    POut.b3 := MIME_ENCODE_TABLE[b and $3F];
    b := b shr 6;
    POut.b2 := MIME_ENCODE_TABLE[b and $3F];
    b := b shr 6;
    POut.b1 := MIME_ENCODE_TABLE[b];

    Inc(POut);
  end;

  case InputByteCount - InMax3 of
    1:
      begin
        b := pIn^;

        b := b shl 4;

        POut.b2 := MIME_ENCODE_TABLE[b and $3F];
        b := b shr 6;
        POut.b1 := MIME_ENCODE_TABLE[b];

        POut.b3 := Ord('='); // Fill remaining 2 bytes.
        POut.b4 := Ord('=');
      end;
    2:
      begin
        b := pIn^;
        Inc(pIn);
        b := b shl 8;
        b := b or pIn^;

        b := b shl 2;

        POut.b3 := MIME_ENCODE_TABLE[b and $3F];
        b := b shr 6;
        POut.b2 := MIME_ENCODE_TABLE[b and $3F];
        b := b shr 6;
        POut.b1 := MIME_ENCODE_TABLE[b];

        POut.b4 := Ord('='); // Fill remaining byte.
      end;
  end;
end;

procedure MimeEncodeStream(const InputStream: TStream; const OutputStream: TStream);
var
  InputBuffer: array[0..BUFFER_SIZE - 1] of Byte;
  OutputBuffer: array[0..((BUFFER_SIZE + 2) div 3) * 4 - 1] of Byte;
  BytesRead: Integer;
begin
  BytesRead := InputStream.Read(InputBuffer, SizeOf(InputBuffer));
  while BytesRead = SizeOf(InputBuffer) do
  begin
    MimeEncode(@InputBuffer, SizeOf(InputBuffer), @OutputBuffer);
    OutputStream.Write(OutputBuffer, SizeOf(OutputBuffer));
    BytesRead := InputStream.Read(InputBuffer, SizeOf(InputBuffer));
  end;
  if BytesRead > 0 then
  begin
    MimeEncode(@InputBuffer, BytesRead, @OutputBuffer);
    OutputStream.Write(OutputBuffer, MimeEncodedSize(BytesRead));
  end;
end;

function MimeDecodedSize(const i: Integer): Integer;
begin
  Result := (i + 3) div 4 * 3;
end;

function MimeDecodePartialEnd(const OutputBuffer: Pointer; const ByteBuffer: Cardinal; const ByteBufferSpace: Cardinal): Integer;
var
  lByteBuffer: Cardinal;
begin
{$IFDEF RangeChecking}
  if OutputBuffer = nil then
    raise EMime.Create(SOutputBufferNil);
{$ENDIF}
  case ByteBufferSpace of
    1:
      begin
        lByteBuffer := ByteBuffer shr 2;
        PByte3(OutputBuffer).b2 := Byte(lByteBuffer);
        lByteBuffer := lByteBuffer shr 8;
        PByte3(OutputBuffer).b1 := Byte(lByteBuffer);
        Result := 2;
      end;
    2:
      begin
        lByteBuffer := ByteBuffer shr 4;
        PByte3(OutputBuffer).b1 := Byte(lByteBuffer);
        Result := 1;
      end;
  else
    Result := 0;
  end;
end;

function MimeDecodePartial(const InputBuffer: Pointer; const InputBytesCount: Integer; const OutputBuffer: Pointer; var ByteBuffer: Cardinal; var ByteBufferSpace: Cardinal): Integer;
var
  lByteBuffer, lByteBufferSpace, c: Cardinal;
  pIn, PInLimit: ^Byte;
  POut: PByte3;
begin
{$IFDEF RangeChecking}
  if InputBuffer = nil then
    raise EMime.Create(SInputBufferNil);
  if OutputBuffer = nil then
    raise EMime.Create(SOutputBufferNil);
{$ENDIF}
  if InputBytesCount > 0 then
  begin
    pIn := InputBuffer;
    Integer(PInLimit) := Integer(pIn) + InputBytesCount;

    POut := OutputBuffer;

    lByteBuffer := ByteBuffer;
    lByteBufferSpace := ByteBufferSpace;

    while pIn <> PInLimit do
    begin
      c := MIME_DECODE_TABLE[pIn^]; // Read from InputBuffer.
      Inc(pIn);

      if c = $FF then
        Continue;

      lByteBuffer := lByteBuffer shl 6;
      lByteBuffer := lByteBuffer or c;
      Dec(lByteBufferSpace);

      if lByteBufferSpace <> 0 then
        Continue; // Read 4 bytes from InputBuffer?

      POut.b3 := Byte(lByteBuffer); // Write 3 bytes to OutputBuffer (in reverse order).
      lByteBuffer := lByteBuffer shr 8;
      POut.b2 := Byte(lByteBuffer);
      lByteBuffer := lByteBuffer shr 8;
      POut.b1 := Byte(lByteBuffer);
      lByteBuffer := 0;
      Inc(POut);

      lByteBufferSpace := 4;
    end;

    ByteBuffer := lByteBuffer;
    ByteBufferSpace := lByteBufferSpace;

    Result := Integer(POut) - Integer(OutputBuffer);
  end
  else
    Result := 0;
end;

procedure MimeDecodeStream(const InputStream: TStream; const OutputStream: TStream);
var
  ByteBuffer, ByteBufferSpace: Cardinal;
  InputBuffer: array[0..BUFFER_SIZE - 1] of Byte;
  OutputBuffer: array[0..(BUFFER_SIZE + 3) div 4 * 3 - 1] of Byte;
  BytesRead: Integer;
begin
  ByteBuffer := 0;
  ByteBufferSpace := 4;
  BytesRead := InputStream.Read(InputBuffer, SizeOf(InputBuffer));
  while BytesRead > 0 do
  begin
    OutputStream.Write(OutputBuffer, MimeDecodePartial(@InputBuffer, BytesRead, @OutputBuffer, ByteBuffer, ByteBufferSpace));
    BytesRead := InputStream.Read(InputBuffer, SizeOf(InputBuffer));
  end;
  OutputStream.Write(OutputBuffer, MimeDecodePartialEnd(@OutputBuffer, ByteBuffer, ByteBufferSpace));
end;

function NumberToken(var SrcStr: WideString; IncludeTokenText: Boolean = False): Integer;
var
  i: Integer;
begin
  Result := 0;

  if SrcStr = '' then
    Exit;

  while True do
  begin
    if not (SrcStr[1] in [WideChar('0')..WideChar('9')]) then
      Delete(SrcStr, 1, 1)
    else
      Break;
  end;

  for i := 1 to Length(SrcStr) do
    if not (SrcStr[i] in [WideChar('0')..WideChar('9')]) then
    begin
      Result := StrToIntDef(Copy(SrcStr, 1, i - 1), 0);
      Delete(SrcStr, 1, i - 1);
      Exit;
    end;

  Result := StrToIntDef(SrcStr, 0);
  SrcStr := '';
end;

function GetTimeZoneBias: Integer;
var
  TimeZoneInfo: TTimeZoneInformation;
begin
  case GetTimeZoneInformation(TimeZoneInfo) of
    TIME_ZONE_ID_STANDARD: Result := TimeZoneInfo.Bias + TimeZoneInfo.StandardBias;
    TIME_ZONE_ID_DAYLIGHT: Result := TimeZoneInfo.Bias + TimeZoneInfo.DaylightBias;
    TIME_ZONE_ID_INVALID: Result := 0;
  else
    Result := TimeZoneInfo.Bias;
  end;
end;

const
  BoolHourMarker: array [Boolean] of string =
      (XMLHourOffsetMinusMarker, XMLHourOffsetPlusMarker);

function FixDateStr(Str: WideString): WideString;
var
  Y, M, D: Word;
begin
  Str := UpperCase(Trim(Str));
  Str := StringReplace(Str, '/', XMLDateSeparator, [rfReplaceAll]);
  Str := StringReplace(Str, '.', XMLDateSeparator, [rfReplaceAll]);
  Str := StringReplace(Str, ',', XMLDateSeparator, [rfReplaceAll]);
  Str := StringReplace(Str, ' ', XMLDateSeparator, [rfReplaceAll]);

  Y := NumberToken(Str);
  M := NumberToken(Str);
  D := NumberToken(Str);
  Result := Format('%.04d%s%.02d%s%.02d%s',
    [Y, XMLDateSeparator, M, XMLDateSeparator, D, Str]);
end;

function FixTimeStr(Str: WideString): WideString;
var
  H, N, S, MS: Word;
  TimeZoneBias: Integer;
begin
  Str := UpperCase(Trim(Str));
  Str := StringReplace(Str, '-', XMLTimeSeparator, [rfReplaceAll]);
  Str := StringReplace(Str, '.', XMLTimeSeparator, [rfReplaceAll]);
  Str := StringReplace(Str, ' ', XMLTimeSeparator, [rfReplaceAll]);

  H := NumberToken(Str);
  N := NumberToken(Str);
  S := NumberToken(Str);
  if (Str <> '') and (Str[1] in [WideChar('.'), WideChar(':')]) then
    MS := NumberToken(Str)
  else
    MS := 0;
  if (Str = '') or not (Str[1] in [WideChar('Z'), WideChar('+'), WideChar('-')])  then
  begin
    TimeZoneBias := GetTimeZoneBias;
    Str := Format('%s%.02d%s%.02d', [BoolHourMarker[TimeZoneBias < 0],
        (TimeZoneBias div 60)*-1, XMLTimeSeparator, (TimeZoneBias mod 60)*-1]);
  end;

  Result := Format('%.02d%s%.02d%s%.02d%s%.03d%s',
    [H, XMLTimeSeparator, N, XMLTimeSeparator, S, SoapDecimalSeparator, MS, Str]);
end;

function LocaleDateTimeToGMTDateTime(const Value: TDateTime): TDateTime;
begin
  Result := Value + (GetTimeZoneBias / MinsPerDay);
end;

function GMTDateTimeToLocaleDateTime(const Value: TDateTime): TDateTime;
begin
  Result := Value - (GetTimeZoneBias / MinsPerDay);
end;

function ExStrToDateTime(const Value: WideString): TDateTime;
var
  TestStr, DateStr, TimeStr: WideString;
begin
  Result := 0;

  TestStr := Trim(Value);

  if TestStr = '' then
    Exit;

  if Pos(' ', TestStr) > 0 then
    TestStr[Pos(' ', TestStr)] := SoapTimePrefix;

  if Pos(SoapTimePrefix, TestStr) = 0 then
  begin
    if Pos(XMLTimeSeparator, TestStr) > 0 then
      Result := ExStrToTime(TestStr)
    else
      Result := ExStrToDate(TestStr);
  end
  else
  begin
    try
      DateStr := Copy(TestStr, 1, Pos(SoapTimePrefix, TestStr) - 1);
      TimeStr := Copy(TestStr, Pos(SoapTimePrefix, TestStr) + 1, MaxInt);

      DateStr := FixDateStr(DateStr);
      TimeStr := FixTimeStr(TimeStr);

      Result := ExStrToDate(DateStr) + ExStrToTime(TimeStr);
    except
      Result := 0;
    end;
  end;
end;

function ExStrToDate(const Value: WideString): TDateTime;
var
  TestStr: WideString;
  XSDate: TXSDate;
begin
  Result := 0;

  try
    TestStr := Trim(Value);

    if TestStr = '' then
      Exit;

    TestStr := FixDateStr(TestStr);

    XSDate := TXSDate.Create;
    try
      XSDate.XSToNative(TestStr);
      Result := EncodeDate(XSDate.Year, XSDate.Month, XSDate.Day);
    finally
      XSDate.Free;
    end;
  except
    Result := 0;
  end;
end;

function ExStrToTime(const Value: WideString): TDateTime;
var
  TestStr: WideString;
  XSTime: TXSTime;
begin
  Result := 0;

  try
    TestStr := Trim(Value);

    if Pos('T', TestStr) > 0 then
      Delete(TestStr, 1, Pos('T', TestStr));

    if TestStr = '' then
      Exit;

    TestStr := FixTimeStr(TestStr);

    XSTime := TXSTime.Create;
    try
      XSTime.UseZeroMilliseconds := False;
      XSTime.XSToNative(TestStr);
      Result := GMTDateTimeToLocaleDateTime(
          EncodeTime(
            (XSTime.Hour-XSTime.HourOffset) mod 24,
            (XSTime.Minute-XSTime.MinuteOffset) mod 60,
            XSTime.Second,
            XSTime.Millisecond)
          );
    finally
      XSTime.Free;
    end;
  except
    Result := 0;
  end;
end;

function ExDateTimeToStr(const Value: TDateTime): WideString;
  function IsOnlyDate: Boolean;
  var
    D1, D2, D3: Word;
  begin
    DecodeDate(Value, D1, D2, D3);
    Result := Value = EncodeDate(D1, D2, D3);
  end;
  function IsOnlyTime: Boolean;
  var
    D1, D2, D3: Word;
  begin
    DecodeDate(Value, D1, D2, D3);
    Result := (D1 = 1899) and (D2 = 12) and (D3 = 30);
  end;
begin
  Result := '';

  if Value = 0 then
    Result := ''
  else
  begin
    if IsOnlyDate then
      Result := ExDateToStr(Value)
    else if IsOnlyTime then
      Result := ExTimeToStr(Value)
    else
      Result := ExDateToStr(Value) + SoapTimePrefix + ExTimeToStr(Value);
  end;
end;

function ExDateToStr(const Value: TDateTime): WideString;
var
  XSDate: TXSDate;
begin
  Result := '';

  if Value = 0 then
    Result := ''
  else
  begin
    XSDate := TXSDate.Create;
    try
      XSDate.AsDate := Value;
      Result := XSDate.NativeToXS;
    finally
      XSDate.Free;
    end;
  end;
end;

function ExTimeToStr(const Value: TDateTime): WideString;
var
  XSTime: TXSTime;
  TimeZoneBias: Integer;
begin
  Result := '';

  if Value = 0 then
    Result := ''
  else
  begin
    XSTime := TXSTime.Create;
    try
      XSTime.UseZeroMilliseconds := False;
      ShowMessage(DateTimeToStr(Value));
      XSTime.AsTime := Value;
      TimeZoneBias := GetTimeZoneBias;
      Result := Format('%.02d%s%.02d%s%.02d%s%.03d%s%.02d%s%.02d',
          [XSTime.Hour, XMLTimeSeparator, XSTime.Minute,
          XMLTimeSeparator, XSTime.Second,
          SoapDecimalSeparator, XSTime.Millisecond, BoolHourMarker[TimeZoneBias < 0],
          Abs(TimeZoneBias div 60), XMLTimeSeparator, Abs(TimeZoneBias mod 60)])
    finally
      XSTime.Free;
    end;
  end;
end;

function Base64ToStr(Base64: WideString): WideString;
var
  S1, S2: TStringStream;
begin
  S2 := TStringStream.Create(Trim(Base64));
  try
    S1 := TStringStream.Create('');
    try
      S2.Position := 0;
      MimeDecodeStream(S2, S1);
      Result := S1.DataString;
    finally
      S1.Free;
    end;
  finally
    S2.Free;
  end;
end;

function StrToBase64(Str: WideString): WideString;
var
  S1, S2: TStringStream;
begin
  S1 := TStringStream.Create(Str);
  try
    S2 := TStringStream.Create('');
    try
      S1.Position := 0;
      MimeEncodeStream(S1, S2);
      Result := S2.DataString;
    finally
      S2.Free;
    end;
  finally
    S1.Free;
  end;
end;

const
  Rcon: array [1..30] of longword = (
    $00000001, $00000002, $00000004, $00000008, $00000010, $00000020,
    $00000040, $00000080, $0000001B, $00000036, $0000006C, $000000D8,
    $000000AB, $0000004D, $0000009A, $0000002F, $0000005E, $000000BC,
    $00000063, $000000C6, $00000097, $00000035, $0000006A, $000000D4,
    $000000B3, $0000007D, $000000FA, $000000EF, $000000C5, $00000091
  );
  
  ForwardTable: array [0..255] of longword = (
    $A56363C6, $847C7CF8, $997777EE, $8D7B7BF6, $0DF2F2FF, $BD6B6BD6, $B16F6FDE, $54C5C591,
    $50303060, $03010102, $A96767CE, $7D2B2B56, $19FEFEE7, $62D7D7B5, $E6ABAB4D, $9A7676EC,
    $45CACA8F, $9D82821F, $40C9C989, $877D7DFA, $15FAFAEF, $EB5959B2, $C947478E, $0BF0F0FB,
    $ECADAD41, $67D4D4B3, $FDA2A25F, $EAAFAF45, $BF9C9C23, $F7A4A453, $967272E4, $5BC0C09B,
    $C2B7B775, $1CFDFDE1, $AE93933D, $6A26264C, $5A36366C, $413F3F7E, $02F7F7F5, $4FCCCC83,
    $5C343468, $F4A5A551, $34E5E5D1, $08F1F1F9, $937171E2, $73D8D8AB, $53313162, $3F15152A,
    $0C040408, $52C7C795, $65232346, $5EC3C39D, $28181830, $A1969637, $0F05050A, $B59A9A2F,
    $0907070E, $36121224, $9B80801B, $3DE2E2DF, $26EBEBCD, $6927274E, $CDB2B27F, $9F7575EA,
    $1B090912, $9E83831D, $742C2C58, $2E1A1A34, $2D1B1B36, $B26E6EDC, $EE5A5AB4, $FBA0A05B,
    $F65252A4, $4D3B3B76, $61D6D6B7, $CEB3B37D, $7B292952, $3EE3E3DD, $712F2F5E, $97848413,
    $F55353A6, $68D1D1B9, $00000000, $2CEDEDC1, $60202040, $1FFCFCE3, $C8B1B179, $ED5B5BB6,
    $BE6A6AD4, $46CBCB8D, $D9BEBE67, $4B393972, $DE4A4A94, $D44C4C98, $E85858B0, $4ACFCF85,
    $6BD0D0BB, $2AEFEFC5, $E5AAAA4F, $16FBFBED, $C5434386, $D74D4D9A, $55333366, $94858511,
    $CF45458A, $10F9F9E9, $06020204, $817F7FFE, $F05050A0, $443C3C78, $BA9F9F25, $E3A8A84B,
    $F35151A2, $FEA3A35D, $C0404080, $8A8F8F05, $AD92923F, $BC9D9D21, $48383870, $04F5F5F1,
    $DFBCBC63, $C1B6B677, $75DADAAF, $63212142, $30101020, $1AFFFFE5, $0EF3F3FD, $6DD2D2BF,
    $4CCDCD81, $140C0C18, $35131326, $2FECECC3, $E15F5FBE, $A2979735, $CC444488, $3917172E,
    $57C4C493, $F2A7A755, $827E7EFC, $473D3D7A, $AC6464C8, $E75D5DBA, $2B191932, $957373E6,
    $A06060C0, $98818119, $D14F4F9E, $7FDCDCA3, $66222244, $7E2A2A54, $AB90903B, $8388880B,
    $CA46468C, $29EEEEC7, $D3B8B86B, $3C141428, $79DEDEA7, $E25E5EBC, $1D0B0B16, $76DBDBAD,
    $3BE0E0DB, $56323264, $4E3A3A74, $1E0A0A14, $DB494992, $0A06060C, $6C242448, $E45C5CB8,
    $5DC2C29F, $6ED3D3BD, $EFACAC43, $A66262C4, $A8919139, $A4959531, $37E4E4D3, $8B7979F2,
    $32E7E7D5, $43C8C88B, $5937376E, $B76D6DDA, $8C8D8D01, $64D5D5B1, $D24E4E9C, $E0A9A949,
    $B46C6CD8, $FA5656AC, $07F4F4F3, $25EAEACF, $AF6565CA, $8E7A7AF4, $E9AEAE47, $18080810,
    $D5BABA6F, $887878F0, $6F25254A, $722E2E5C, $241C1C38, $F1A6A657, $C7B4B473, $51C6C697,
    $23E8E8CB, $7CDDDDA1, $9C7474E8, $211F1F3E, $DD4B4B96, $DCBDBD61, $868B8B0D, $858A8A0F,
    $907070E0, $423E3E7C, $C4B5B571, $AA6666CC, $D8484890, $05030306, $01F6F6F7, $120E0E1C,
    $A36161C2, $5F35356A, $F95757AE, $D0B9B969, $91868617, $58C1C199, $271D1D3A, $B99E9E27,
    $38E1E1D9, $13F8F8EB, $B398982B, $33111122, $BB6969D2, $70D9D9A9, $898E8E07, $A7949433,
    $B69B9B2D, $221E1E3C, $92878715, $20E9E9C9, $49CECE87, $FF5555AA, $78282850, $7ADFDFA5,
    $8F8C8C03, $F8A1A159, $80898909, $170D0D1A, $DABFBF65, $31E6E6D7, $C6424284, $B86868D0,
    $C3414182, $B0999929, $772D2D5A, $110F0F1E, $CBB0B07B, $FC5454A8, $D6BBBB6D, $3A16162C
  );
  
  LastForwardTable: array [0..255] of longword = (
    $00000063, $0000007C, $00000077, $0000007B, $000000F2, $0000006B, $0000006F, $000000C5,
    $00000030, $00000001, $00000067, $0000002B, $000000FE, $000000D7, $000000AB, $00000076,
    $000000CA, $00000082, $000000C9, $0000007D, $000000FA, $00000059, $00000047, $000000F0,
    $000000AD, $000000D4, $000000A2, $000000AF, $0000009C, $000000A4, $00000072, $000000C0,
    $000000B7, $000000FD, $00000093, $00000026, $00000036, $0000003F, $000000F7, $000000CC,
    $00000034, $000000A5, $000000E5, $000000F1, $00000071, $000000D8, $00000031, $00000015,
    $00000004, $000000C7, $00000023, $000000C3, $00000018, $00000096, $00000005, $0000009A,
    $00000007, $00000012, $00000080, $000000E2, $000000EB, $00000027, $000000B2, $00000075,
    $00000009, $00000083, $0000002C, $0000001A, $0000001B, $0000006E, $0000005A, $000000A0,
    $00000052, $0000003B, $000000D6, $000000B3, $00000029, $000000E3, $0000002F, $00000084,
    $00000053, $000000D1, $00000000, $000000ED, $00000020, $000000FC, $000000B1, $0000005B,
    $0000006A, $000000CB, $000000BE, $00000039, $0000004A, $0000004C, $00000058, $000000CF,
    $000000D0, $000000EF, $000000AA, $000000FB, $00000043, $0000004D, $00000033, $00000085,
    $00000045, $000000F9, $00000002, $0000007F, $00000050, $0000003C, $0000009F, $000000A8,
    $00000051, $000000A3, $00000040, $0000008F, $00000092, $0000009D, $00000038, $000000F5,
    $000000BC, $000000B6, $000000DA, $00000021, $00000010, $000000FF, $000000F3, $000000D2,
    $000000CD, $0000000C, $00000013, $000000EC, $0000005F, $00000097, $00000044, $00000017,
    $000000C4, $000000A7, $0000007E, $0000003D, $00000064, $0000005D, $00000019, $00000073,
    $00000060, $00000081, $0000004F, $000000DC, $00000022, $0000002A, $00000090, $00000088,
    $00000046, $000000EE, $000000B8, $00000014, $000000DE, $0000005E, $0000000B, $000000DB,
    $000000E0, $00000032, $0000003A, $0000000A, $00000049, $00000006, $00000024, $0000005C,
    $000000C2, $000000D3, $000000AC, $00000062, $00000091, $00000095, $000000E4, $00000079,
    $000000E7, $000000C8, $00000037, $0000006D, $0000008D, $000000D5, $0000004E, $000000A9,
    $0000006C, $00000056, $000000F4, $000000EA, $00000065, $0000007A, $000000AE, $00000008,
    $000000BA, $00000078, $00000025, $0000002E, $0000001C, $000000A6, $000000B4, $000000C6,
    $000000E8, $000000DD, $00000074, $0000001F, $0000004B, $000000BD, $0000008B, $0000008A,
    $00000070, $0000003E, $000000B5, $00000066, $00000048, $00000003, $000000F6, $0000000E,
    $00000061, $00000035, $00000057, $000000B9, $00000086, $000000C1, $0000001D, $0000009E,
    $000000E1, $000000F8, $00000098, $00000011, $00000069, $000000D9, $0000008E, $00000094,
    $0000009B, $0000001E, $00000087, $000000E9, $000000CE, $00000055, $00000028, $000000DF,
    $0000008C, $000000A1, $00000089, $0000000D, $000000BF, $000000E6, $00000042, $00000068,
    $00000041, $00000099, $0000002D, $0000000F, $000000B0, $00000054, $000000BB, $00000016
  );

  InverseTable: array [0..255] of longword = (
    $50A7F451, $5365417E, $C3A4171A, $965E273A, $CB6BAB3B, $F1459D1F, $AB58FAAC, $9303E34B,
    $55FA3020, $F66D76AD, $9176CC88, $254C02F5, $FCD7E54F, $D7CB2AC5, $80443526, $8FA362B5,
    $495AB1DE, $671BBA25, $980EEA45, $E1C0FE5D, $02752FC3, $12F04C81, $A397468D, $C6F9D36B,
    $E75F8F03, $959C9215, $EB7A6DBF, $DA595295, $2D83BED4, $D3217458, $2969E049, $44C8C98E,
    $6A89C275, $78798EF4, $6B3E5899, $DD71B927, $B64FE1BE, $17AD88F0, $66AC20C9, $B43ACE7D,
    $184ADF63, $82311AE5, $60335197, $457F5362, $E07764B1, $84AE6BBB, $1CA081FE, $942B08F9,
    $58684870, $19FD458F, $876CDE94, $B7F87B52, $23D373AB, $E2024B72, $578F1FE3, $2AAB5566,
    $0728EBB2, $03C2B52F, $9A7BC586, $A50837D3, $F2872830, $B2A5BF23, $BA6A0302, $5C8216ED,
    $2B1CCF8A, $92B479A7, $F0F207F3, $A1E2694E, $CDF4DA65, $D5BE0506, $1F6234D1, $8AFEA6C4,
    $9D532E34, $A055F3A2, $32E18A05, $75EBF6A4, $39EC830B, $AAEF6040, $069F715E, $51106EBD,
    $F98A213E, $3D06DD96, $AE053EDD, $46BDE64D, $B58D5491, $055DC471, $6FD40604, $FF155060,
    $24FB9819, $97E9BDD6, $CC434089, $779ED967, $BD42E8B0, $888B8907, $385B19E7, $DBEEC879,
    $470A7CA1, $E90F427C, $C91E84F8, $00000000, $83868009, $48ED2B32, $AC70111E, $4E725A6C,
    $FBFF0EFD, $5638850F, $1ED5AE3D, $27392D36, $64D90F0A, $21A65C68, $D1545B9B, $3A2E3624,
    $B1670A0C, $0FE75793, $D296EEB4, $9E919B1B, $4FC5C080, $A220DC61, $694B775A, $161A121C,
    $0ABA93E2, $E52AA0C0, $43E0223C, $1D171B12, $0B0D090E, $ADC78BF2, $B9A8B62D, $C8A91E14,
    $8519F157, $4C0775AF, $BBDD99EE, $FD607FA3, $9F2601F7, $BCF5725C, $C53B6644, $347EFB5B,
    $7629438B, $DCC623CB, $68FCEDB6, $63F1E4B8, $CADC31D7, $10856342, $40229713, $2011C684,
    $7D244A85, $F83DBBD2, $1132F9AE, $6DA129C7, $4B2F9E1D, $F330B2DC, $EC52860D, $D0E3C177,
    $6C16B32B, $99B970A9, $FA489411, $2264E947, $C48CFCA8, $1A3FF0A0, $D82C7D56, $EF903322,
    $C74E4987, $C1D138D9, $FEA2CA8C, $360BD498, $CF81F5A6, $28DE7AA5, $268EB7DA, $A4BFAD3F,
    $E49D3A2C, $0D927850, $9BCC5F6A, $62467E54, $C2138DF6, $E8B8D890, $5EF7392E, $F5AFC382,
    $BE805D9F, $7C93D069, $A92DD56F, $B31225CF, $3B99ACC8, $A77D1810, $6E639CE8, $7BBB3BDB,
    $097826CD, $F418596E, $01B79AEC, $A89A4F83, $656E95E6, $7EE6FFAA, $08CFBC21, $E6E815EF,
    $D99BE7BA, $CE366F4A, $D4099FEA, $D67CB029, $AFB2A431, $31233F2A, $3094A5C6, $C066A235,
    $37BC4E74, $A6CA82FC, $B0D090E0, $15D8A733, $4A9804F1, $F7DAEC41, $0E50CD7F, $2FF69117,
    $8DD64D76, $4DB0EF43, $544DAACC, $DF0496E4, $E3B5D19E, $1B886A4C, $B81F2CC1, $7F516546,
    $04EA5E9D, $5D358C01, $737487FA, $2E410BFB, $5A1D67B3, $52D2DB92, $335610E9, $1347D66D,
    $8C61D79A, $7A0CA137, $8E14F859, $893C13EB, $EE27A9CE, $35C961B7, $EDE51CE1, $3CB1477A,
    $59DFD29C, $3F73F255, $79CE1418, $BF37C773, $EACDF753, $5BAAFD5F, $146F3DDF, $86DB4478,
    $81F3AFCA, $3EC468B9, $2C342438, $5F40A3C2, $72C31D16, $0C25E2BC, $8B493C28, $41950DFF,
    $7101A839, $DEB30C08, $9CE4B4D8, $90C15664, $6184CB7B, $70B632D5, $745C6C48, $4257B8D0
  );

  LastInverseTable: array [0..255] of longword = (
    $00000052, $00000009, $0000006A, $000000D5, $00000030, $00000036, $000000A5, $00000038,
    $000000BF, $00000040, $000000A3, $0000009E, $00000081, $000000F3, $000000D7, $000000FB,
    $0000007C, $000000E3, $00000039, $00000082, $0000009B, $0000002F, $000000FF, $00000087,
    $00000034, $0000008E, $00000043, $00000044, $000000C4, $000000DE, $000000E9, $000000CB,
    $00000054, $0000007B, $00000094, $00000032, $000000A6, $000000C2, $00000023, $0000003D,
    $000000EE, $0000004C, $00000095, $0000000B, $00000042, $000000FA, $000000C3, $0000004E,
    $00000008, $0000002E, $000000A1, $00000066, $00000028, $000000D9, $00000024, $000000B2,
    $00000076, $0000005B, $000000A2, $00000049, $0000006D, $0000008B, $000000D1, $00000025,
    $00000072, $000000F8, $000000F6, $00000064, $00000086, $00000068, $00000098, $00000016,
    $000000D4, $000000A4, $0000005C, $000000CC, $0000005D, $00000065, $000000B6, $00000092,
    $0000006C, $00000070, $00000048, $00000050, $000000FD, $000000ED, $000000B9, $000000DA,
    $0000005E, $00000015, $00000046, $00000057, $000000A7, $0000008D, $0000009D, $00000084,
    $00000090, $000000D8, $000000AB, $00000000, $0000008C, $000000BC, $000000D3, $0000000A,
    $000000F7, $000000E4, $00000058, $00000005, $000000B8, $000000B3, $00000045, $00000006,
    $000000D0, $0000002C, $0000001E, $0000008F, $000000CA, $0000003F, $0000000F, $00000002,
    $000000C1, $000000AF, $000000BD, $00000003, $00000001, $00000013, $0000008A, $0000006B,
    $0000003A, $00000091, $00000011, $00000041, $0000004F, $00000067, $000000DC, $000000EA,
    $00000097, $000000F2, $000000CF, $000000CE, $000000F0, $000000B4, $000000E6, $00000073,
    $00000096, $000000AC, $00000074, $00000022, $000000E7, $000000AD, $00000035, $00000085,
    $000000E2, $000000F9, $00000037, $000000E8, $0000001C, $00000075, $000000DF, $0000006E,
    $00000047, $000000F1, $0000001A, $00000071, $0000001D, $00000029, $000000C5, $00000089,
    $0000006F, $000000B7, $00000062, $0000000E, $000000AA, $00000018, $000000BE, $0000001B,
    $000000FC, $00000056, $0000003E, $0000004B, $000000C6, $000000D2, $00000079, $00000020,
    $0000009A, $000000DB, $000000C0, $000000FE, $00000078, $000000CD, $0000005A, $000000F4,
    $0000001F, $000000DD, $000000A8, $00000033, $00000088, $00000007, $000000C7, $00000031,
    $000000B1, $00000012, $00000010, $00000059, $00000027, $00000080, $000000EC, $0000005F,
    $00000060, $00000051, $0000007F, $000000A9, $00000019, $000000B5, $0000004A, $0000000D,
    $0000002D, $000000E5, $0000007A, $0000009F, $00000093, $000000C9, $0000009C, $000000EF,
    $000000A0, $000000E0, $0000003B, $0000004D, $000000AE, $0000002A, $000000F5, $000000B0,
    $000000C8, $000000EB, $000000BB, $0000003C, $00000083, $00000053, $00000099, $00000061,
    $00000017, $0000002B, $00000004, $0000007E, $000000BA, $00000077, $000000D6, $00000026,
    $000000E1, $00000069, $00000014, $00000063, $00000055, $00000021, $0000000C, $0000007D
  );

type
  TAESBuffer = array [0..15] of byte;
  TAESKey128 = array [0..15] of byte;
  TAESExpandedKey128 = array [0..43] of longword;

resourcestring
  SInvalidInBufSize = 'Invalid buffer size for decryption';
  SReadError = 'Stream read error';
  SWriteError = 'Stream write error';
  
function Min(A, B: integer): integer;
begin
  if A < B then
    Result := A
  else
    Result := B;
end;

procedure ExpandAESKeyForEncryption(const Key: TAESKey128; var ExpandedKey: TAESExpandedKey128);
var
  I, J: integer;
  T: longword;
  W0, W1, W2, W3: longword;
begin
  ExpandedKey[0] := PLongWord(@Key[0])^;
  ExpandedKey[1] := PLongWord(@Key[4])^;
  ExpandedKey[2] := PLongWord(@Key[8])^;
  ExpandedKey[3] := PLongWord(@Key[12])^;
  I := 0; J := 1;
  repeat
    T := (ExpandedKey[I + 3] shl 24) or (ExpandedKey[I + 3] shr 8);
    W0 := LastForwardTable[Byte(T)]; W1 := LastForwardTable[Byte(T shr 8)];
    W2 := LastForwardTable[Byte(T shr 16)]; W3 := LastForwardTable[Byte(T shr 24)];
    ExpandedKey[I + 4] := ExpandedKey[I] xor
      (W0 xor ((W1 shl 8) or (W1 shr 24)) xor
      ((W2 shl 16) or (W2 shr 16)) xor ((W3 shl 24) or (W3 shr 8))) xor Rcon[J];
    Inc(J);
    ExpandedKey[I + 5] := ExpandedKey[I + 1] xor ExpandedKey[I + 4];
    ExpandedKey[I + 6] := ExpandedKey[I + 2] xor ExpandedKey[I + 5];
    ExpandedKey[I + 7] := ExpandedKey[I + 3] xor ExpandedKey[I + 6];
    Inc(I, 4);
  until I >= 40;
end;

procedure EncryptAES(const InBuf: TAESBuffer; const Key: TAESExpandedKey128;
  var OutBuf: TAESBuffer); 
var
  T0, T1: array [0..3] of longword;
  W0, W1, W2, W3: longword;
begin
  // initializing
  T0[0] := PLongWord(@InBuf[0])^ xor Key[0];
  T0[1] := PLongWord(@InBuf[4])^ xor Key[1];
  T0[2] := PLongWord(@InBuf[8])^ xor Key[2];
  T0[3] := PLongWord(@InBuf[12])^ xor Key[3];
  // performing transformation 9 times
  // round 1
  W0 := ForwardTable[Byte(T0[0])]; W1 := ForwardTable[Byte(T0[1] shr 8)];
  W2 := ForwardTable[Byte(T0[2] shr 16)]; W3 := ForwardTable[Byte(T0[3] shr 24)];
  T1[0] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[4];
  W0 := ForwardTable[Byte(T0[1])]; W1 := ForwardTable[Byte(T0[2] shr 8)];
  W2 := ForwardTable[Byte(T0[3] shr 16)]; W3 := ForwardTable[Byte(T0[0] shr 24)];
  T1[1] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[5];
  W0 := ForwardTable[Byte(T0[2])]; W1 := ForwardTable[Byte(T0[3] shr 8)];
  W2 := ForwardTable[Byte(T0[0] shr 16)]; W3 := ForwardTable[Byte(T0[1] shr 24)];
  T1[2] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[6];
  W0 := ForwardTable[Byte(T0[3])]; W1 := ForwardTable[Byte(T0[0] shr 8)];
  W2 := ForwardTable[Byte(T0[1] shr 16)]; W3 := ForwardTable[Byte(T0[2] shr 24)];
  T1[3] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[7];
  // round 2
  W0 := ForwardTable[Byte(T1[0])]; W1 := ForwardTable[Byte(T1[1] shr 8)];
  W2 := ForwardTable[Byte(T1[2] shr 16)]; W3 := ForwardTable[Byte(T1[3] shr 24)];
  T0[0] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[8];
  W0 := ForwardTable[Byte(T1[1])]; W1 := ForwardTable[Byte(T1[2] shr 8)];
  W2 := ForwardTable[Byte(T1[3] shr 16)]; W3 := ForwardTable[Byte(T1[0] shr 24)];
  T0[1] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[9];
  W0 := ForwardTable[Byte(T1[2])]; W1 := ForwardTable[Byte(T1[3] shr 8)];
  W2 := ForwardTable[Byte(T1[0] shr 16)]; W3 := ForwardTable[Byte(T1[1] shr 24)];
  T0[2] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[10];
  W0 := ForwardTable[Byte(T1[3])]; W1 := ForwardTable[Byte(T1[0] shr 8)];
  W2 := ForwardTable[Byte(T1[1] shr 16)]; W3 := ForwardTable[Byte(T1[2] shr 24)];
  T0[3] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[11];
  // round 3
  W0 := ForwardTable[Byte(T0[0])]; W1 := ForwardTable[Byte(T0[1] shr 8)];
  W2 := ForwardTable[Byte(T0[2] shr 16)]; W3 := ForwardTable[Byte(T0[3] shr 24)];
  T1[0] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[12];
  W0 := ForwardTable[Byte(T0[1])]; W1 := ForwardTable[Byte(T0[2] shr 8)];
  W2 := ForwardTable[Byte(T0[3] shr 16)]; W3 := ForwardTable[Byte(T0[0] shr 24)];
  T1[1] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[13];
  W0 := ForwardTable[Byte(T0[2])]; W1 := ForwardTable[Byte(T0[3] shr 8)];
  W2 := ForwardTable[Byte(T0[0] shr 16)]; W3 := ForwardTable[Byte(T0[1] shr 24)];
  T1[2] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[14];
  W0 := ForwardTable[Byte(T0[3])]; W1 := ForwardTable[Byte(T0[0] shr 8)];
  W2 := ForwardTable[Byte(T0[1] shr 16)]; W3 := ForwardTable[Byte(T0[2] shr 24)];
  T1[3] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[15];
  // round 4
  W0 := ForwardTable[Byte(T1[0])]; W1 := ForwardTable[Byte(T1[1] shr 8)];
  W2 := ForwardTable[Byte(T1[2] shr 16)]; W3 := ForwardTable[Byte(T1[3] shr 24)];
  T0[0] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[16];
  W0 := ForwardTable[Byte(T1[1])]; W1 := ForwardTable[Byte(T1[2] shr 8)];
  W2 := ForwardTable[Byte(T1[3] shr 16)]; W3 := ForwardTable[Byte(T1[0] shr 24)];
  T0[1] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[17];
  W0 := ForwardTable[Byte(T1[2])]; W1 := ForwardTable[Byte(T1[3] shr 8)];
  W2 := ForwardTable[Byte(T1[0] shr 16)]; W3 := ForwardTable[Byte(T1[1] shr 24)];
  T0[2] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[18];
  W0 := ForwardTable[Byte(T1[3])]; W1 := ForwardTable[Byte(T1[0] shr 8)];
  W2 := ForwardTable[Byte(T1[1] shr 16)]; W3 := ForwardTable[Byte(T1[2] shr 24)];
  T0[3] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[19];
  // round 5
  W0 := ForwardTable[Byte(T0[0])]; W1 := ForwardTable[Byte(T0[1] shr 8)];
  W2 := ForwardTable[Byte(T0[2] shr 16)]; W3 := ForwardTable[Byte(T0[3] shr 24)];
  T1[0] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[20];
  W0 := ForwardTable[Byte(T0[1])]; W1 := ForwardTable[Byte(T0[2] shr 8)];
  W2 := ForwardTable[Byte(T0[3] shr 16)]; W3 := ForwardTable[Byte(T0[0] shr 24)];
  T1[1] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[21];
  W0 := ForwardTable[Byte(T0[2])]; W1 := ForwardTable[Byte(T0[3] shr 8)];
  W2 := ForwardTable[Byte(T0[0] shr 16)]; W3 := ForwardTable[Byte(T0[1] shr 24)];
  T1[2] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[22];
  W0 := ForwardTable[Byte(T0[3])]; W1 := ForwardTable[Byte(T0[0] shr 8)];
  W2 := ForwardTable[Byte(T0[1] shr 16)]; W3 := ForwardTable[Byte(T0[2] shr 24)];
  T1[3] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[23];
  // round 6
  W0 := ForwardTable[Byte(T1[0])]; W1 := ForwardTable[Byte(T1[1] shr 8)];
  W2 := ForwardTable[Byte(T1[2] shr 16)]; W3 := ForwardTable[Byte(T1[3] shr 24)];
  T0[0] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[24];
  W0 := ForwardTable[Byte(T1[1])]; W1 := ForwardTable[Byte(T1[2] shr 8)];
  W2 := ForwardTable[Byte(T1[3] shr 16)]; W3 := ForwardTable[Byte(T1[0] shr 24)];
  T0[1] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[25];
  W0 := ForwardTable[Byte(T1[2])]; W1 := ForwardTable[Byte(T1[3] shr 8)];
  W2 := ForwardTable[Byte(T1[0] shr 16)]; W3 := ForwardTable[Byte(T1[1] shr 24)];
  T0[2] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[26];
  W0 := ForwardTable[Byte(T1[3])]; W1 := ForwardTable[Byte(T1[0] shr 8)];
  W2 := ForwardTable[Byte(T1[1] shr 16)]; W3 := ForwardTable[Byte(T1[2] shr 24)];
  T0[3] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[27];
  // round 7
  W0 := ForwardTable[Byte(T0[0])]; W1 := ForwardTable[Byte(T0[1] shr 8)];
  W2 := ForwardTable[Byte(T0[2] shr 16)]; W3 := ForwardTable[Byte(T0[3] shr 24)];
  T1[0] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[28];
  W0 := ForwardTable[Byte(T0[1])]; W1 := ForwardTable[Byte(T0[2] shr 8)];
  W2 := ForwardTable[Byte(T0[3] shr 16)]; W3 := ForwardTable[Byte(T0[0] shr 24)];
  T1[1] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[29];
  W0 := ForwardTable[Byte(T0[2])]; W1 := ForwardTable[Byte(T0[3] shr 8)];
  W2 := ForwardTable[Byte(T0[0] shr 16)]; W3 := ForwardTable[Byte(T0[1] shr 24)];
  T1[2] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[30];
  W0 := ForwardTable[Byte(T0[3])]; W1 := ForwardTable[Byte(T0[0] shr 8)];
  W2 := ForwardTable[Byte(T0[1] shr 16)]; W3 := ForwardTable[Byte(T0[2] shr 24)];
  T1[3] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[31];
  // round 8
  W0 := ForwardTable[Byte(T1[0])]; W1 := ForwardTable[Byte(T1[1] shr 8)];
  W2 := ForwardTable[Byte(T1[2] shr 16)]; W3 := ForwardTable[Byte(T1[3] shr 24)];
  T0[0] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[32];
  W0 := ForwardTable[Byte(T1[1])]; W1 := ForwardTable[Byte(T1[2] shr 8)];
  W2 := ForwardTable[Byte(T1[3] shr 16)]; W3 := ForwardTable[Byte(T1[0] shr 24)];
  T0[1] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[33];
  W0 := ForwardTable[Byte(T1[2])]; W1 := ForwardTable[Byte(T1[3] shr 8)];
  W2 := ForwardTable[Byte(T1[0] shr 16)]; W3 := ForwardTable[Byte(T1[1] shr 24)];
  T0[2] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[34];
  W0 := ForwardTable[Byte(T1[3])]; W1 := ForwardTable[Byte(T1[0] shr 8)];
  W2 := ForwardTable[Byte(T1[1] shr 16)]; W3 := ForwardTable[Byte(T1[2] shr 24)];
  T0[3] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[35];
  // round 9
  W0 := ForwardTable[Byte(T0[0])]; W1 := ForwardTable[Byte(T0[1] shr 8)];
  W2 := ForwardTable[Byte(T0[2] shr 16)]; W3 := ForwardTable[Byte(T0[3] shr 24)];
  T1[0] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[36];
  W0 := ForwardTable[Byte(T0[1])]; W1 := ForwardTable[Byte(T0[2] shr 8)];
  W2 := ForwardTable[Byte(T0[3] shr 16)]; W3 := ForwardTable[Byte(T0[0] shr 24)];
  T1[1] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[37];
  W0 := ForwardTable[Byte(T0[2])]; W1 := ForwardTable[Byte(T0[3] shr 8)];
  W2 := ForwardTable[Byte(T0[0] shr 16)]; W3 := ForwardTable[Byte(T0[1] shr 24)];
  T1[2] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[38];
  W0 := ForwardTable[Byte(T0[3])]; W1 := ForwardTable[Byte(T0[0] shr 8)];
  W2 := ForwardTable[Byte(T0[1] shr 16)]; W3 := ForwardTable[Byte(T0[2] shr 24)];
  T1[3] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[39];
  // last round of transformations
  W0 := LastForwardTable[Byte(T1[0])]; W1 := LastForwardTable[Byte(T1[1] shr 8)];
  W2 := LastForwardTable[Byte(T1[2] shr 16)]; W3 := LastForwardTable[Byte(T1[3] shr 24)];
  T0[0] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[40];
  W0 := LastForwardTable[Byte(T1[1])]; W1 := LastForwardTable[Byte(T1[2] shr 8)];
  W2 := LastForwardTable[Byte(T1[3] shr 16)]; W3 := LastForwardTable[Byte(T1[0] shr 24)];
  T0[1] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[41];
  W0 := LastForwardTable[Byte(T1[2])]; W1 := LastForwardTable[Byte(T1[3] shr 8)];
  W2 := LastForwardTable[Byte(T1[0] shr 16)]; W3 := LastForwardTable[Byte(T1[1] shr 24)];
  T0[2] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[42];
  W0 := LastForwardTable[Byte(T1[3])]; W1 := LastForwardTable[Byte(T1[0] shr 8)];
  W2 := LastForwardTable[Byte(T1[1] shr 16)]; W3 := LastForwardTable[Byte(T1[2] shr 24)];
  T0[3] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[43];
  // finalizing
  PLongWord(@OutBuf[0])^ := T0[0]; PLongWord(@OutBuf[4])^ := T0[1];
  PLongWord(@OutBuf[8])^ := T0[2]; PLongWord(@OutBuf[12])^ := T0[3];
end;

procedure EncryptAESStreamECB(Source: TStream; Count: cardinal;
  const ExpandedKey: TAESExpandedKey128; Dest: TStream); overload;
var
  TempIn, TempOut: TAESBuffer;
  Done: cardinal;
begin
  if Count = 0 then
  begin
    Source.Position := 0;
    Count := Source.Size;
  end
  else Count := Min(Count, Source.Size - Source.Position);
  if Count = 0 then exit;
  while Count >= SizeOf(TAESBuffer) do
  begin
    Done := Source.Read(TempIn, SizeOf(TempIn));
    if Done < SizeOf(TempIn) then
      raise EStreamError.Create(SReadError);
    EncryptAES(TempIn, ExpandedKey, TempOut);
    Done := Dest.Write(TempOut, SizeOf(TempOut));
    if Done < SizeOf(TempOut) then
      raise EStreamError.Create(SWriteError);
    Dec(Count, SizeOf(TAESBuffer));
  end;
  if Count > 0 then
  begin
    Done := Source.Read(TempIn, Count);
    if Done < Count then
      raise EStreamError.Create(SReadError);
    FillChar(TempIn[Count], SizeOf(TempIn) - Count, 0);
    EncryptAES(TempIn, ExpandedKey, TempOut);
    Done := Dest.Write(TempOut, SizeOf(TempOut));
    if Done < SizeOf(TempOut) then
      raise EStreamError.Create(SWriteError);
  end;
end;

procedure EncryptAESStreamECB(Source: TStream; Count: cardinal;
  const Key: TAESKey128; Dest: TStream); overload;
var
  ExpandedKey: TAESExpandedKey128;
begin
  ExpandAESKeyForEncryption(Key, ExpandedKey);
  EncryptAESStreamECB(Source, Count, ExpandedKey, Dest);
end;

procedure ExpandAESKeyForDecryption(var ExpandedKey: TAESExpandedKey128); overload;
var
  I: integer;
  U, F2, F4, F8, F9: longword;
begin
  for I := 1 to 9 do
  begin
    F9 := ExpandedKey[I * 4];
    U := F9 and $80808080;
    F2 := ((F9 and $7F7F7F7F) shl 1) xor ((U - (U shr 7)) and $1B1B1B1B);
    U := F2 and $80808080;
    F4 := ((F2 and $7F7F7F7F) shl 1) xor ((U - (U shr 7)) and $1B1B1B1B);
    U := F4 and $80808080;
    F8 := ((F4 and $7F7F7F7F) shl 1) xor ((U - (U shr 7)) and $1B1B1B1B);
    F9 := F9 xor F8;
    ExpandedKey[I * 4] := F2 xor F4 xor F8 xor
      (((F2 xor F9) shl 24) or ((F2 xor F9) shr 8)) xor
      (((F4 xor F9) shl 16) or ((F4 xor F9) shr 16)) xor ((F9 shl 8) or (F9 shr 24));
    F9 := ExpandedKey[I * 4 + 1];
    U := F9 and $80808080;
    F2 := ((F9 and $7F7F7F7F) shl 1) xor ((U - (U shr 7)) and $1B1B1B1B);
    U := F2 and $80808080;
    F4 := ((F2 and $7F7F7F7F) shl 1) xor ((U - (U shr 7)) and $1B1B1B1B);
    U := F4 and $80808080;
    F8 := ((F4 and $7F7F7F7F) shl 1) xor ((U - (U shr 7)) and $1B1B1B1B);
    F9 := F9 xor F8;
    ExpandedKey[I * 4 + 1] := F2 xor F4 xor F8 xor
      (((F2 xor F9) shl 24) or ((F2 xor F9) shr 8)) xor
      (((F4 xor F9) shl 16) or ((F4 xor F9) shr 16)) xor ((F9 shl 8) or (F9 shr 24));
    F9 := ExpandedKey[I * 4 + 2];
    U := F9 and $80808080;
    F2 := ((F9 and $7F7F7F7F) shl 1) xor ((U - (U shr 7)) and $1B1B1B1B);
    U := F2 and $80808080;
    F4 := ((F2 and $7F7F7F7F) shl 1) xor ((U - (U shr 7)) and $1B1B1B1B);
    U := F4 and $80808080;
    F8 := ((F4 and $7F7F7F7F) shl 1) xor ((U - (U shr 7)) and $1B1B1B1B);
    F9 := F9 xor F8;
    ExpandedKey[I * 4 + 2] := F2 xor F4 xor F8 xor
      (((F2 xor F9) shl 24) or ((F2 xor F9) shr 8)) xor
      (((F4 xor F9) shl 16) or ((F4 xor F9) shr 16)) xor ((F9 shl 8) or (F9 shr 24));
    F9 := ExpandedKey[I * 4 + 3];
    U := F9 and $80808080;
    F2 := ((F9 and $7F7F7F7F) shl 1) xor ((U - (U shr 7)) and $1B1B1B1B);
    U := F2 and $80808080;
    F4 := ((F2 and $7F7F7F7F) shl 1) xor ((U - (U shr 7)) and $1B1B1B1B);
    U := F4 and $80808080;
    F8 := ((F4 and $7F7F7F7F) shl 1) xor ((U - (U shr 7)) and $1B1B1B1B);
    F9 := F9 xor F8;
    ExpandedKey[I * 4 + 3] := F2 xor F4 xor F8 xor
      (((F2 xor F9) shl 24) or ((F2 xor F9) shr 8)) xor
      (((F4 xor F9) shl 16) or ((F4 xor F9) shr 16)) xor ((F9 shl 8) or (F9 shr 24));
  end;
end;

procedure ExpandAESKeyForDecryption(const Key: TAESKey128; var ExpandedKey: TAESExpandedKey128); overload;
begin
  ExpandAESKeyForEncryption(Key, ExpandedKey);
  ExpandAESKeyForDecryption(ExpandedKey);
end;

procedure DecryptAES(const InBuf: TAESBuffer; const Key: TAESExpandedKey128;
  var OutBuf: TAESBuffer);
var
  T0, T1: array [0..3] of longword;
  W0, W1, W2, W3: longword;
begin
  // initializing
  T0[0] := PLongWord(@InBuf[0])^ xor Key[40];
  T0[1] := PLongWord(@InBuf[4])^ xor Key[41];
  T0[2] := PLongWord(@InBuf[8])^ xor Key[42];
  T0[3] := PLongWord(@InBuf[12])^ xor Key[43];
  // performing transformations 9 times
  // round 1
  W0 := InverseTable[Byte(T0[0])]; W1 := InverseTable[Byte(T0[3] shr 8)];
  W2 := InverseTable[Byte(T0[2] shr 16)]; W3 := InverseTable[Byte(T0[1] shr 24)];
  T1[0] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[36];
  W0 := InverseTable[Byte(T0[1])]; W1 := InverseTable[Byte(T0[0] shr 8)];
  W2 := InverseTable[Byte(T0[3] shr 16)]; W3 := InverseTable[Byte(T0[2] shr 24)];
  T1[1] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[37];
  W0 := InverseTable[Byte(T0[2])]; W1 := InverseTable[Byte(T0[1] shr 8)];
  W2 := InverseTable[Byte(T0[0] shr 16)]; W3 := InverseTable[Byte(T0[3] shr 24)];
  T1[2] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[38];
  W0 := InverseTable[Byte(T0[3])]; W1 := InverseTable[Byte(T0[2] shr 8)];
  W2 := InverseTable[Byte(T0[1] shr 16)]; W3 := InverseTable[Byte(T0[0] shr 24)];
  T1[3] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[39];
  // round 2
  W0 := InverseTable[Byte(T1[0])]; W1 := InverseTable[Byte(T1[3] shr 8)];
  W2 := InverseTable[Byte(T1[2] shr 16)]; W3 := InverseTable[Byte(T1[1] shr 24)];
  T0[0] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[32];
  W0 := InverseTable[Byte(T1[1])]; W1 := InverseTable[Byte(T1[0] shr 8)];
  W2 := InverseTable[Byte(T1[3] shr 16)]; W3 := InverseTable[Byte(T1[2] shr 24)];
  T0[1] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[33];
  W0 := InverseTable[Byte(T1[2])]; W1 := InverseTable[Byte(T1[1] shr 8)];
  W2 := InverseTable[Byte(T1[0] shr 16)]; W3 := InverseTable[Byte(T1[3] shr 24)];
  T0[2] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[34];
  W0 := InverseTable[Byte(T1[3])]; W1 := InverseTable[Byte(T1[2] shr 8)];
  W2 := InverseTable[Byte(T1[1] shr 16)]; W3 := InverseTable[Byte(T1[0] shr 24)];
  T0[3] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[35];
  // round 3
  W0 := InverseTable[Byte(T0[0])]; W1 := InverseTable[Byte(T0[3] shr 8)];
  W2 := InverseTable[Byte(T0[2] shr 16)]; W3 := InverseTable[Byte(T0[1] shr 24)];
  T1[0] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[28];
  W0 := InverseTable[Byte(T0[1])]; W1 := InverseTable[Byte(T0[0] shr 8)];
  W2 := InverseTable[Byte(T0[3] shr 16)]; W3 := InverseTable[Byte(T0[2] shr 24)];
  T1[1] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[29];
  W0 := InverseTable[Byte(T0[2])]; W1 := InverseTable[Byte(T0[1] shr 8)];
  W2 := InverseTable[Byte(T0[0] shr 16)]; W3 := InverseTable[Byte(T0[3] shr 24)];
  T1[2] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[30];
  W0 := InverseTable[Byte(T0[3])]; W1 := InverseTable[Byte(T0[2] shr 8)];
  W2 := InverseTable[Byte(T0[1] shr 16)]; W3 := InverseTable[Byte(T0[0] shr 24)];
  T1[3] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[31];
  // round 4
  W0 := InverseTable[Byte(T1[0])]; W1 := InverseTable[Byte(T1[3] shr 8)];
  W2 := InverseTable[Byte(T1[2] shr 16)]; W3 := InverseTable[Byte(T1[1] shr 24)];
  T0[0] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[24];
  W0 := InverseTable[Byte(T1[1])]; W1 := InverseTable[Byte(T1[0] shr 8)];
  W2 := InverseTable[Byte(T1[3] shr 16)]; W3 := InverseTable[Byte(T1[2] shr 24)];
  T0[1] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[25];
  W0 := InverseTable[Byte(T1[2])]; W1 := InverseTable[Byte(T1[1] shr 8)];
  W2 := InverseTable[Byte(T1[0] shr 16)]; W3 := InverseTable[Byte(T1[3] shr 24)];
  T0[2] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[26];
  W0 := InverseTable[Byte(T1[3])]; W1 := InverseTable[Byte(T1[2] shr 8)];
  W2 := InverseTable[Byte(T1[1] shr 16)]; W3 := InverseTable[Byte(T1[0] shr 24)];
  T0[3] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[27];
  // round 5
  W0 := InverseTable[Byte(T0[0])]; W1 := InverseTable[Byte(T0[3] shr 8)];
  W2 := InverseTable[Byte(T0[2] shr 16)]; W3 := InverseTable[Byte(T0[1] shr 24)];
  T1[0] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[20];
  W0 := InverseTable[Byte(T0[1])]; W1 := InverseTable[Byte(T0[0] shr 8)];
  W2 := InverseTable[Byte(T0[3] shr 16)]; W3 := InverseTable[Byte(T0[2] shr 24)];
  T1[1] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[21];
  W0 := InverseTable[Byte(T0[2])]; W1 := InverseTable[Byte(T0[1] shr 8)];
  W2 := InverseTable[Byte(T0[0] shr 16)]; W3 := InverseTable[Byte(T0[3] shr 24)];
  T1[2] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[22];
  W0 := InverseTable[Byte(T0[3])]; W1 := InverseTable[Byte(T0[2] shr 8)];
  W2 := InverseTable[Byte(T0[1] shr 16)]; W3 := InverseTable[Byte(T0[0] shr 24)];
  T1[3] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[23];
  // round 6
  W0 := InverseTable[Byte(T1[0])]; W1 := InverseTable[Byte(T1[3] shr 8)];
  W2 := InverseTable[Byte(T1[2] shr 16)]; W3 := InverseTable[Byte(T1[1] shr 24)];
  T0[0] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[16];
  W0 := InverseTable[Byte(T1[1])]; W1 := InverseTable[Byte(T1[0] shr 8)];
  W2 := InverseTable[Byte(T1[3] shr 16)]; W3 := InverseTable[Byte(T1[2] shr 24)];
  T0[1] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[17];
  W0 := InverseTable[Byte(T1[2])]; W1 := InverseTable[Byte(T1[1] shr 8)];
  W2 := InverseTable[Byte(T1[0] shr 16)]; W3 := InverseTable[Byte(T1[3] shr 24)];
  T0[2] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[18];
  W0 := InverseTable[Byte(T1[3])]; W1 := InverseTable[Byte(T1[2] shr 8)];
  W2 := InverseTable[Byte(T1[1] shr 16)]; W3 := InverseTable[Byte(T1[0] shr 24)];
  T0[3] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[19];
  // round 7
  W0 := InverseTable[Byte(T0[0])]; W1 := InverseTable[Byte(T0[3] shr 8)];
  W2 := InverseTable[Byte(T0[2] shr 16)]; W3 := InverseTable[Byte(T0[1] shr 24)];
  T1[0] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[12];
  W0 := InverseTable[Byte(T0[1])]; W1 := InverseTable[Byte(T0[0] shr 8)];
  W2 := InverseTable[Byte(T0[3] shr 16)]; W3 := InverseTable[Byte(T0[2] shr 24)];
  T1[1] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[13];
  W0 := InverseTable[Byte(T0[2])]; W1 := InverseTable[Byte(T0[1] shr 8)];
  W2 := InverseTable[Byte(T0[0] shr 16)]; W3 := InverseTable[Byte(T0[3] shr 24)];
  T1[2] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[14];
  W0 := InverseTable[Byte(T0[3])]; W1 := InverseTable[Byte(T0[2] shr 8)];
  W2 := InverseTable[Byte(T0[1] shr 16)]; W3 := InverseTable[Byte(T0[0] shr 24)];
  T1[3] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[15];
  // round 8
  W0 := InverseTable[Byte(T1[0])]; W1 := InverseTable[Byte(T1[3] shr 8)];
  W2 := InverseTable[Byte(T1[2] shr 16)]; W3 := InverseTable[Byte(T1[1] shr 24)];
  T0[0] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[8];
  W0 := InverseTable[Byte(T1[1])]; W1 := InverseTable[Byte(T1[0] shr 8)];
  W2 := InverseTable[Byte(T1[3] shr 16)]; W3 := InverseTable[Byte(T1[2] shr 24)];
  T0[1] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[9];
  W0 := InverseTable[Byte(T1[2])]; W1 := InverseTable[Byte(T1[1] shr 8)];
  W2 := InverseTable[Byte(T1[0] shr 16)]; W3 := InverseTable[Byte(T1[3] shr 24)];
  T0[2] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[10];
  W0 := InverseTable[Byte(T1[3])]; W1 := InverseTable[Byte(T1[2] shr 8)];
  W2 := InverseTable[Byte(T1[1] shr 16)]; W3 := InverseTable[Byte(T1[0] shr 24)];
  T0[3] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[11];
  // round 9
  W0 := InverseTable[Byte(T0[0])]; W1 := InverseTable[Byte(T0[3] shr 8)];
  W2 := InverseTable[Byte(T0[2] shr 16)]; W3 := InverseTable[Byte(T0[1] shr 24)];
  T1[0] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[4];
  W0 := InverseTable[Byte(T0[1])]; W1 := InverseTable[Byte(T0[0] shr 8)];
  W2 := InverseTable[Byte(T0[3] shr 16)]; W3 := InverseTable[Byte(T0[2] shr 24)];
  T1[1] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[5];
  W0 := InverseTable[Byte(T0[2])]; W1 := InverseTable[Byte(T0[1] shr 8)];
  W2 := InverseTable[Byte(T0[0] shr 16)]; W3 := InverseTable[Byte(T0[3] shr 24)];
  T1[2] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[6];
  W0 := InverseTable[Byte(T0[3])]; W1 := InverseTable[Byte(T0[2] shr 8)];
  W2 := InverseTable[Byte(T0[1] shr 16)]; W3 := InverseTable[Byte(T0[0] shr 24)];
  T1[3] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[7];
  // last round of transformations
  W0 := LastInverseTable[Byte(T1[0])]; W1 := LastInverseTable[Byte(T1[3] shr 8)];
  W2 := LastInverseTable[Byte(T1[2] shr 16)]; W3 := LastInverseTable[Byte(T1[1] shr 24)];
  T0[0] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[0];
  W0 := LastInverseTable[Byte(T1[1])]; W1 := LastInverseTable[Byte(T1[0] shr 8)];
  W2 := LastInverseTable[Byte(T1[3] shr 16)]; W3 := LastInverseTable[Byte(T1[2] shr 24)];
  T0[1] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[1];
  W0 := LastInverseTable[Byte(T1[2])]; W1 := LastInverseTable[Byte(T1[1] shr 8)];
  W2 := LastInverseTable[Byte(T1[0] shr 16)]; W3 := LastInverseTable[Byte(T1[3] shr 24)];
  T0[2] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[2];
  W0 := LastInverseTable[Byte(T1[3])]; W1 := LastInverseTable[Byte(T1[2] shr 8)];
  W2 := LastInverseTable[Byte(T1[1] shr 16)]; W3 := LastInverseTable[Byte(T1[0] shr 24)];
  T0[3] := (W0 xor ((W1 shl 8) or (W1 shr 24)) xor ((W2 shl 16) or (W2 shr 16))
    xor ((W3 shl 24) or (W3 shr 8))) xor Key[3];
  // finalizing
  PLongWord(@OutBuf[0])^ := T0[0]; PLongWord(@OutBuf[4])^ := T0[1];
  PLongWord(@OutBuf[8])^ := T0[2]; PLongWord(@OutBuf[12])^ := T0[3];
end;

procedure DecryptAESStreamECB(Source: TStream; Count: cardinal;
  const ExpandedKey: TAESExpandedKey128; Dest: TStream); overload;
var
  TempIn, TempOut: TAESBuffer;
  Done: cardinal;
begin
  if Count = 0 then
  begin
    Source.Position := 0;
    Count := Source.Size;
  end
  else Count := Min(Count, Source.Size - Source.Position);
  if Count = 0 then exit;
  if (Count mod SizeOf(TAESBuffer)) > 0 then
    raise Exception.Create(SInvalidInBufSize);
  while Count >= SizeOf(TAESBuffer) do
  begin
    Done := Source.Read(TempIn, SizeOf(TempIn));
    if Done < SizeOf(TempIn) then
      raise EStreamError.Create(SReadError);
    DecryptAES(TempIn, ExpandedKey, TempOut);
    Done := Dest.Write(TempOut, SizeOf(TempOut));
    if Done < SizeOf(TempOut) then
      raise EStreamError.Create(SWriteError);
    Dec(Count, SizeOf(TAESBuffer));
  end;
end;

procedure DecryptAESStreamECB(Source: TStream; Count: cardinal;
  const Key: TAESKey128; Dest: TStream); overload;
var
  ExpandedKey: TAESExpandedKey128;
begin
  ExpandAESKeyForDecryption(Key, ExpandedKey);
  DecryptAESStreamECB(Source, Count, ExpandedKey, Dest);
end;

function RandomStr10Base(StrLen: Integer): string;
var
  Ch, LastCh: Char;
begin
  Result := '';
  LastCh := ' ';
  while Length(Result) < StrLen do
  begin
    Ch := Char(Random($0A)+$30);
    if Ch <> LastCh then
    begin
      Result := Result + Ch;
      LastCh := Ch;
    end;
  end;
end;

function RandomStr256Base(StrLen: Integer): string;
var
  Ch, LastCh: Char;
begin
  Result := '';
  LastCh := ' ';
  while Length(Result) < StrLen do
  begin
    Ch := Char(Random($5C)+$21);
    if Ch <> LastCh then
    begin
      Result := Result + Ch;
      LastCh := Ch;
    end;
  end;
end;

type
  THandledObject = class(TObject)
  protected
    FHandle: THandle;
  public
    destructor Destroy; override;
    property Handle: THandle read FHandle;
  end;

  TMutex = class(THandledObject)
  public
    constructor Create(const Name: string);
    function Get(TimeOut: Integer = -1): Boolean;
    function Release: Boolean;
  end;

{ THandledObject }

destructor THandledObject.Destroy;
begin
  if FHandle <> 0 then
    CloseHandle(FHandle);
end;

{ TMutex }

constructor TMutex.Create(const Name: string);
begin
  FHandle := CreateMutex(nil, False, PChar(Name));
  if FHandle = 0 then
    abort;
end;

function TMutex.Get(TimeOut: Integer = -1): Boolean;
begin
  Result := WaitForSingleObject(FHandle, TimeOut) = WAIT_OBJECT_0;
end;

function TMutex.Release: Boolean;
begin
  Result := ReleaseMutex(FHandle);
end;

{ TXmlIni }

constructor TXmlIni.Create(AFileName: WideString);
begin
  Randomize;

  FList := TList.Create;
  FXmlEncoding := xietUTF_16;
  FAutoSave := False;
  FFileName := '';
  FCaseSensitive := False;
  FCompressed := False;
  FEncryptMode := xiemAuto;
  FPassword := '';
  {$IFDEF RSA_ENCRYPT_ENABLED}
  FPrivateKey := TStringList.Create;
  FPublicKey := TStringList.Create;
  {$ENDIF}
  FRaiseException := False;

  try
    LoadFromFile(AFileName);
  except
  end;
  FFileName := AFileName;

  InitializeCriticalSection(FLockCriSec);
end;

constructor TXmlIni.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  CoInitialize(nil);

  Randomize;

  FList := TList.Create;
  FXmlEncoding := xietUTF_16;
  FAutoSave := False;
  FFileName := '';
  FCaseSensitive := False;
  FCompressed := False;
  FEncryptMode := xiemAuto;
  FPassword := '';
  {$IFDEF RSA_ENCRYPT_ENABLED}
  FPrivateKey := TStringList.Create;
  FPublicKey := TStringList.Create;
  {$ENDIF}
  FRaiseException := False;

  InitializeCriticalSection(FLockCriSec);
end;

destructor TXmlIni.Destroy;
begin
  DeleteCriticalSection(FLockCriSec);

  if FAutoSave then
    SaveToFile;

  Clear;

  {$IFDEF RSA_ENCRYPT_ENABLED}
  FPrivateKey.Free;
  FPublicKey.Free;
  {$ENDIF}

  FList.Free;
end;

function TXmlIni.Equals(Dest: TXmlIni): Boolean;
var
  i: Integer;
begin
  Result := False;

  for i := 0 to Count - 1 do
  begin
    if not Dest.Exists(SectionByIndex[i].Name) then
      Exit;
    if not SectionByIndex[i].Equals(Dest[SectionByIndex[i].Name]) then
      Exit;
  end;

  Result := True;
end;

procedure TXmlIni.CopyFrom(Source: TXmlIni);
begin
  Clear;
  Append(Source);
end;

procedure TXmlIni.Append(Source: TXmlIni);
var
  i: Integer;
begin
  Source.Collation;
  for i := 0 to Source.Count - 1 do
    Add(Source.SectionByIndex[i].Name).Append(Source.SectionByIndex[i]);
  Collation;
end;

function TXmlIni.Add(SectionName: WideString): TXmlIniSection;
begin
  Result := InternalFind(SectionName);
  if Result = nil then
    Result := InternalAdd(SectionName);
end;

procedure TXmlIni.Delete(SectionName: WideString);
var
  i: Integer;
begin
  for i := Count - 1 downto 0 do
    if SectionByIndex[i].Name = SectionName then
    begin
      SectionByIndex[i].Free;
      FList.Delete(i);
      Break;
    end;
end;

function TXmlIni.Exists(SectionName: WideString; KeyName: WideString = ''): Boolean;
var
  Section: TXmlIniSection;
begin
  Result := False;

  Section := InternalFind(SectionName);
  if Section <> nil then
  begin
    if KeyName <> '' then
      Result := Section.Exists(KeyName)
    else
      Result := True;
  end;
end;

procedure TXmlIni.Clear;
var
  i: Integer;
begin
  FXmlEncoding := xietUTF_16;

  for i := Count - 1 downto 0 do
  begin
    SectionByIndex[i].Free;
    FList.Delete(i);
  end;
end;

procedure TXmlIni.Collation;
var
  i: Integer;
begin
  for i := Count - 1 downto 0 do
  begin
    SectionByIndex[i].Collation;
    if SectionByIndex[i].Count = 0 then
    begin
      SectionByIndex[i].Free;
      FList.Delete(i);
    end;
  end;
end;

procedure TXmlIni.LoadFromStream(Stream: TStream);
var
  TestBuf: array[0..1] of Char;
  FullPassword: string;
  Key128: TAESKey128;
  DecryptMS: TMemoryStream;
  UnzipMS: TMemoryStream;
  procedure LoadAESDecrypt;
  begin
    DecryptMS := TMemoryStream.Create;
    try
      DecryptAESStreamECB(Stream, Stream.Size-2, Key128, DecryptMS);

      DecryptMS.Position := 0;

      UnzipMS := TMemoryStream.Create;
      try
        DecompressStream(DecryptMS, UnzipMS);
        InternalXmlLoadFromStream(UnzipMS);
      finally
        UnzipMS.Free;
      end;
    finally
      DecryptMS.Free;
    end;
  end;
{$IFDEF RSA_ENCRYPT_ENABLED}
var
  Key1, Key2, NilGInt: TFGInt;
  KeySize: Integer;
{$ENDIF}
begin
  try
    Stream.Position := 0;
    Stream.Read(TestBuf, 2);

    {$IFDEF RSA_ENCRYPT_ENABLED}
    if TestBuf = HeaderRSACrypto then
    begin
      if FPrivateKey.Text = '' then
        raise Exception.Create('Data encrypted, PrivateKey required.');

      FCompressed := True;
      Stream.Position := 2;

      Stream.Read(KeySize, SizeOf(KeySize));
      FullPassword := StringOfChar(' ', KeySize);
      Stream.Read(PChar(FullPassword)^, KeySize);

      ExtractKey(FPrivateKey.Text, Key1, Key2);
      RSADecrypt(FullPassword, Key1, Key2, NilGInt, NilGInt, NilGInt, NilGInt, FullPassword);

      FullPassword := Copy(FullPassword, FakePasswordLength+1, MaxInt);
      FillChar(Key128, SizeOf(Key128), 0);
      Move(PChar(FullPassword)^, Key128, SizeOf(Key128));

      LoadAESDecrypt;
    end
    else if TestBuf = HeaderAESCrypto then
    {$ELSE}
    if TestBuf = HeaderAESCrypto then
    {$ENDIF}
    begin
      if FPassword = '' then
        raise Exception.Create('Data encrypted, Password required.');

      FCompressed := True;
      Stream.Position := 2;

      FullPassword := FPassword;
      while Length(FullPassword) < SizeOf(Key128) do
        FullPassword := FullPassword + FPassword;
        
      FillChar(Key128, SizeOf(Key128), 0);
      Move(PChar(FullPassword)^, Key128, SizeOf(Key128));

      LoadAESDecrypt;
    end
    else if TestBuf = HeaderCompress then
    begin
      FCompressed := True;
      Stream.Position := 0;
      UnzipMS := TMemoryStream.Create;
      try
        DecompressStream(Stream, UnzipMS);
        InternalXmlLoadFromStream(UnzipMS);
      finally
        UnzipMS.Free;
      end;
    end
    else
    begin
      FCompressed := False;
      Stream.Position := 0;
      InternalXmlLoadFromStream(Stream);
    end;
  except
    if FRaiseException then
      raise;
  end;
end;

procedure TXmlIni.SaveToStream(Stream: TStream);
var
  FullPassword: string;
  Key128: TAESKey128;
  EncryptMS: TMemoryStream;
  ZipMS: TMemoryStream;
  CurEncMode: TXmlIniEncryptMode;
  procedure SaveAESEncrypt;
  begin
    EncryptMS := TMemoryStream.Create;
    try
      ZipMS := TMemoryStream.Create;
      try
        InternalXmlSaveToStream(ZipMS);
        ZipMS.Position := 0;
        EncryptMS.Position := 0;
        CompressStream(ZipMS, EncryptMS);
      finally
        ZipMS.Free;
      end;

      EncryptMS.Position := 0;
      EncryptAESStreamECB(EncryptMS, 0, Key128, Stream);
    finally
      EncryptMS.Free;
    end;
  end;
{$IFDEF RSA_ENCRYPT_ENABLED}
var
  KeySize: Integer;
  Key1, Key2: TFGInt;
{$ENDIF}
begin
  try
    CurEncMode := FEncryptMode;

    if CurEncMode = xiemAuto then
    begin
{$IFDEF RSA_ENCRYPT_ENABLED}
      if FPublicKey.Text <> '' then
        CurEncMode := xiemRSA
      else
{$ENDIF}
      if FPassword <> '' then
        CurEncMode := xiemAES
      else
        CurEncMode := xiemPlain;
    end;

    case CurEncMode of
{$IFDEF RSA_ENCRYPT_ENABLED}
      xiemRSA:
        begin
          FCompressed := True;

          FullPassword := RandomStr256Base(SizeOf(Key128));
          Move(PChar(FullPassword)^, Key128, SizeOf(Key128));

          FullPassword := RandomStr256Base(FakePasswordLength) + FullPassword;

          ExtractKey(FPublicKey.Text, Key1, Key2);

          RSAEncrypt(FullPassword, Key1, Key2, FullPassword);
          KeySize := Length(FullPassword);

          Stream.Position := 0;
          Stream.Write(HeaderRSACrypto, 2);
          Stream.Write(KeySize, SizeOf(KeySize));
          Stream.Write(PChar(FullPassword)^, KeySize);

          SaveAESEncrypt;
        end;
{$ENDIF}
      xiemAES:
        begin
          FCompressed := True;

          FullPassword := FPassword;
          while Length(FullPassword) < SizeOf(Key128) do
            FullPassword := FullPassword + FPassword;

          FillChar(Key128, SizeOf(Key128), 0);
          Move(PChar(FullPassword)^, Key128, SizeOf(Key128));

          Stream.Position := 0;
          Stream.Write(HeaderAESCrypto, 2);
          
          SaveAESEncrypt;
        end;
      xiemPlain:
        begin
          if FCompressed then
          begin
            ZipMS := TMemoryStream.Create;
            try
              InternalXmlSaveToStream(ZipMS);
              ZipMS.Position := 0;
              Stream.Position := 0;
              CompressStream(ZipMS, Stream);
            finally
              ZipMS.Free;
            end;
          end
          else
            InternalXmlSaveToStream(Stream);
        end;
    end;
  except
    if FRaiseException then
      raise;
  end;
end;

procedure TXmlIni.LoadFromString(Value: WideString);
var
  SS: TStringStream;
begin
  if TrimLeft(Value) = '' then
    Clear
  else if TrimLeft(Value)[1] = '<' then
  begin
    SS := TStringStream.Create(Value);
    try
      LoadFromStream(SS);
    finally
      SS.Free;
    end;
  end
  else
    LoadFromBase64String(Value);
end;

function TXmlIni.SaveToString: WideString;
var
  OldXIEncoding: TXmlIniEncodingType;
  MS: TMemoryStream;
  NullChar: WideChar;
begin
  if not FCompressed then
  begin
    OldXIEncoding := XmlEncoding;
    try
      XmlEncoding := xietUTF_16;

      MS := TMemoryStream.Create;
      try
        SaveToStream(MS);
        NullChar := #0;
        MS.Write(NullChar, 2);
        Result := PWideChar(PChar(MS.Memory) + 2);
      finally
        MS.Free;
      end;
    finally
      XmlEncoding := OldXIEncoding;
    end;
  end
  else
    Result := SaveToBase64String;
end;

procedure TXmlIni.LoadFromBase64String(Value: string);
var
  SS: TStringStream;
  MS: TMemoryStream;
begin
  if (TrimLeft(Value) = '') or (TrimLeft(Value)[1] = '<') then
    LoadFromString(Value)
  else
  begin
    SS := TStringStream.Create(Trim(Value));
    try
      MS := TMemoryStream.Create;
      try
        SS.Position := 0;
        MimeDecodeStream(SS, MS);
        LoadFromStream(MS);
      finally
        MS.Free;
      end;
    finally
      SS.Free;
    end;
  end;
end;

function TXmlIni.SaveToBase64String: string;
var
  SS: TStringStream;
  MS: TMemoryStream;
begin
  MS := TMemoryStream.Create;
  try
    SaveToStream(MS);

    SS := TStringStream.Create('');
    try
      MS.Position := 0;
      MimeEncodeStream(MS, SS);
      Result := SS.DataString;
    finally
      SS.Free;
    end;
  finally
    MS.Free;
  end;
end;

procedure TXmlIni.LoadFromFile(AFileName: WideString = '');
begin
  AFileName := Trim(AFileName);

  if (AFileName <> '') and (ExtractFilePath(AFileName) = '') then
    AFileName := ExtractFilePath(ParamStr(0)) + AFileName;

  if AFileName = '' then
    AFileName := FFileName;

  if AFileName = '' then
    Exit;

  if not FileExists(AFileName) then
    if FRaiseException then
      raise Exception.Create('File "' + AFileName + '" not exists.')
    else
      Exit;

  FFileName := AFileName;

  InternalLoadFromFile(FFileName);
end;

procedure TXmlIni.SaveToFile(AFileName: WideString = '');
begin
  AFileName := Trim(AFileName);

  if (AFileName <> '') and (ExtractFilePath(AFileName) = '') then
    AFileName := ExtractFilePath(ParamStr(0)) + AFileName;

  if AFileName = '' then
    AFileName := FFileName;

  if AFileName = '' then
    if FRaiseException then
      raise Exception.Create('File name is required.')
    else
      Exit;

  FFileName := AFileName;

  InternalSaveToFile(FFileName);
end;

procedure TXmlIni.LoadFromRegistry(RootKey: HKEY; Path: WideString);
var
  Reg: TRegistry;
  Names: TStringList;
  i: Integer;
begin
  Clear;

  if (Path = '') or (Path[Length(Path)] <> '\') then
    Path := Path + '\';

  Reg := TRegistry.Create;
  try
    Reg.RootKey := RootKey;

    if Reg.OpenKey(Path, False) then
    begin
      Names := TStringList.Create;
      try
        Reg.GetKeyNames(Names);
        for i := 0 to Names.Count - 1 do
          Section[Names[i]].LoadFromRegistry(RootKey, Path, Names[i]);

        Section[''].LoadFromRegistry(RootKey, Path);

        Collation;
      finally
        Names.Free;
      end;
    end;
  finally
    Reg.Free;
  end;
end;

procedure TXmlIni.SaveToRegistry(RootKey: HKEY; Path: WideString);
var
  i: Integer;
begin
  Collation;

  if (Path = '') or (Path[Length(Path)] <> '\') then
    Path := Path + '\';

  for i := 0 to Count - 1 do
    SectionByIndex[i].SaveToRegistry(RootKey, Path, SectionByIndex[i].Name);
end;

procedure TXmlIni.LoadFromIni(AIniFileName: WideString);
  function SingleLineToText(Str: WideString): WideString;
  begin
    Result := Str;

    if Pos('&', Result) > 0 then
    begin
      Result := StringReplace(Result, '&#xD;', #13, [rfIgnoreCase, rfReplaceAll]);
      Result := StringReplace(Result, '&#xA;', #10, [rfIgnoreCase, rfReplaceAll]);
      Result := StringReplace(Result, '&amp;', '&', [rfIgnoreCase, rfReplaceAll]);
      Result := StringReplace(Result, '&', '&amp;', [rfReplaceAll]);
    end;
  end;
var
  Ini: TIniFile;
  SectionNames, KeyNameValues: TStringList;
  SecName, KeyName, KeyValue: WideString;
  XISection: TXmlIniSection;
  i, j: Integer;
begin
  Clear;

  AIniFileName := Trim(AIniFileName);

  if (AIniFileName <> '') and (ExtractFilePath(AIniFileName) = '') then
    AIniFileName := ExtractFilePath(ParamStr(0)) + AIniFileName;

  if (AIniFileName = '') or not FileExists(AIniFileName) then
    if FRaiseException then
      raise Exception.Create('File "' + AIniFileName + '" not exists.')
    else
      Exit;

  Ini := TIniFile.Create(AIniFileName);
  try
    SectionNames := TStringList.Create;
    KeyNameValues := TStringList.Create;
    try
      SectionNames.Duplicates := dupIgnore;
      KeyNameValues.Duplicates := dupIgnore;

      Ini.ReadSections(SectionNames);
      
      for i := 0 to SectionNames.Count - 1 do
      begin
        SecName := SingleLineToText(SectionNames[i]);
        XISection := InternalAdd(SecName);

        Ini.ReadSectionValues(SectionNames[i], KeyNameValues);

        for j := 0 to KeyNameValues.Count - 1 do
        begin
          KeyValue := KeyNameValues.ValueFromIndex[j];
          if KeyValue <> '' then
          begin
            KeyName := SingleLineToText(KeyNameValues.Names[j]);
            XISection.InternalAdd(KeyName).Text := SingleLineToText(KeyValue);
          end;
        end;
      end;

      Collation;
    finally
      KeyNameValues.Free;
      SectionNames.Free;
    end;
  finally
    Ini.Free;
  end;
end;

procedure TXmlIni.SaveToIni(AIniFileName: WideString);
  function TextToSingleLine(Str: WideString): WideString;
  begin
    Result := Str;
    if (Pos(#13, Result) > 0) or (Pos(#10, Result) > 0) then
    begin
      Result := StringReplace(Result, '&', '&amp;', [rfReplaceAll]);
      Result := StringReplace(Result, #10, '&#xA;', [rfReplaceAll]);
      Result := StringReplace(Result, #13, '&#xD;', [rfReplaceAll]);
    end;
  end;
var
  Ini: TIniFile;
  i, j: Integer;
begin
  Collation;

  AIniFileName := Trim(AIniFileName);

  if (AIniFileName <> '') and (ExtractFilePath(AIniFileName) = '') then
    AIniFileName := ExtractFilePath(ParamStr(0)) + AIniFileName;

  if AIniFileName = '' then
    if FRaiseException then
      raise Exception.Create('File name is required.')
    else
      Exit;

  DeleteFile(AIniFileName);

  Ini := TIniFile.Create(AIniFileName);
  try
    for i := 0 to Count - 1 do
      for j := 0 to SectionByIndex[i].Count - 1 do
        Ini.WriteString(
          TextToSingleLine(SectionByIndex[i].Name),
          TextToSingleLine(SectionByIndex[i].KeyByIndex[j].Name),
          TextToSingleLine(SectionByIndex[i].KeyByIndex[j].Text));
  finally
    Ini.Free;
  end;
end;

type
  TSetMemMemoryStream = class(TMemoryStream)
  public
    procedure SetMem(Ptr: Pointer; Size: Longint);
  end;

procedure TSetMemMemoryStream.SetMem(Ptr: Pointer; Size: Integer);
begin
  SetPointer(Ptr, Size);
end;

procedure TXmlIni.LoadFromShareMem(ShareMemID: string = '');
var
  MemLockMutex: TMutex;
  MS: TSetMemMemoryStream;
  ShareMemHandle: THandle;
  ShareMemMap: Pointer;
  ShareMemSize: Integer;
begin
  ShareMemID := Trim(ShareMemID);

  if ShareMemID = '' then
    Exit;

  ShareMemID := XI_ShareName + ShareMemID;

  MemLockMutex := TMutex.Create(XI_LockMutexName + ShareMemID);
  try
    if MemLockMutex.Get then
    begin
      try
        try
          ShareMemHandle := CreateFileMapping($FFFFFFFF, nil, PAGE_READONLY, 0,
            SizeOf(Integer), PChar(ShareMemID));

          ShareMemMap := MapViewOfFile(ShareMemHandle, FILE_MAP_READ, 0, 0, SizeOf(Integer));
          try
            Move(ShareMemMap^, ShareMemSize, SizeOf(Integer));
          finally
            UnmapViewOfFile(ShareMemMap);
          end;

          if ShareMemSize = 0 then
            Clear
          else
          begin
            ShareMemHandle := CreateFileMapping($FFFFFFFF, nil, PAGE_READONLY, 0,
              SizeOf(Integer) + ShareMemSize, PChar(ShareMemID));

            ShareMemMap := MapViewOfFile(ShareMemHandle, FILE_MAP_READ, 0, 0, SizeOf(Integer) + ShareMemSize);
            try
              MS := TSetMemMemoryStream.Create;
              try
                MS.SetMem(Pointer(PChar(ShareMemMap) + SizeOf(Integer)), ShareMemSize);
                LoadFromStream(MS);
                MS.SetMem(nil, 0);
              finally
                MS.Free;
              end;
            finally
              UnmapViewOfFile(ShareMemMap);
            end;
          end;
        except
        end;
      finally
        MemLockMutex.Release;
      end;
    end;
  finally
    MemLockMutex.Free;
  end;
end;

procedure TXmlIni.SaveToShareMem(ShareMemID: string = '');
var
  MemLockMutex: TMutex;
  MS: TMemoryStream;
  ShareMemHandle: THandle;
  ShareMemMap: Pointer;
  ShareMemSize: Integer;
begin
  ShareMemID := Trim(ShareMemID);

  if ShareMemID = '' then
    if FRaiseException then
      raise Exception.Create('ShareMemID is required.')
    else
      Exit;

  ShareMemID := XI_ShareName + ShareMemID;

  MemLockMutex := TMutex.Create(XI_LockMutexName + ShareMemID);
  try
    if MemLockMutex.Get then
    begin
      try
        try
          MS := TMemoryStream.Create;
          try
            SaveToStream(MS);

            ShareMemHandle := CreateFileMapping($FFFFFFFF, nil, PAGE_READWRITE, 0,
              SizeOf(Integer) + MS.Size, PChar(ShareMemID));

            ShareMemMap := MapViewOfFile(ShareMemHandle, FILE_MAP_WRITE, 0, 0, SizeOf(Integer) + MS.Size);
            try
              ShareMemSize := MS.Size;
              Move(ShareMemSize, ShareMemMap^, SizeOf(Integer));
              Move(MS.Memory^, Pointer(PChar(ShareMemMap) + SizeOf(Integer))^, MS.Size);
            finally
              UnmapViewOfFile(ShareMemMap);
            end;
          finally
            MS.Free;
          end;
        except
        end;
      finally
        MemLockMutex.Release;
      end;
    end;
  finally
    MemLockMutex.Free;
  end;
end;

{$IFDEF RSA_ENCRYPT_ENABLED}
procedure TXmlIni.GenerateNewKey;
  function KeyToStr(Key1, Key2: TFGInt): string;
  var
    InMS: TMemoryStream;
    OutSS: TStringStream;
    i: Integer;
    KSize: ShortInt;
  begin
    InMS := TMemoryStream.Create;
    try
      KSize := Key1.Number[0];
      if Key1.Sign = Negative then
        KSize := KSize * -1;
      InMS.Write(KSize, SizeOf(KSize));
      for i := 1 to Length(Key1.Number)-1 do
        InMS.Write(Key1.Number[i], SizeOf(LongWord));

      KSize := Key2.Number[0];
      if Key2.Sign = Negative then
        KSize := KSize * -1;
      InMS.Write(KSize, SizeOf(KSize));
      for i := 1 to Length(Key2.Number)-1 do
        InMS.Write(Key2.Number[i], SizeOf(LongWord));

      OutSS := TStringStream.Create('');
      try
        InMS.Position := 0;
        MimeEncodeStream(InMS, OutSS);
        Result := OutSS.DataString;
      finally
        OutSS.Free;
      end;
    finally
      InMS.Free;
    end;
  end;
var
  n, e, d, dp, dq, p, q, phi, one, two, gcd, temp, nilgint: TFGInt;
  GenCount: Integer;
begin
  Base10StringToFGInt(RandomStr10Base(32), p);
  PrimeSearch(p);
  Base256StringToFGInt(RandomStr256Base(32), q);
  PrimeSearch(q);
  FGIntMul(p, q, n);
  p.Number[1] := p.Number[1] - 1;
  q.Number[1] := q.Number[1] - 1;
  FGIntMul(p, q, phi);
  Base10StringToFGInt(RandomStr10Base(32), e);
  Base10StringToFGInt('1', one);
  Base10StringToFGInt('2', two);
  
  FGIntGCD(phi, e, gcd);
  GenCount := 0;
  while FGIntCompareAbs(gcd, one) <> Eq do
  begin
    FGIntadd(e, two, temp);
    FGIntCopy(temp, e);
    FGIntGCD(phi, e, gcd);
    Inc(GenCount);
    if GenCount > 10 then
    begin
      Base10StringToFGInt(RandomStr10Base(32), e);
      FGIntGCD(phi, e, gcd);
    end;
  end;
  FGIntDestroy(two);
  FGIntDestroy(one);
  FGIntDestroy(gcd);
  FGIntModInv(e, phi, d);
  FGIntModInv(e, p, dp);
  FGIntModInv(e, q, dq);

  FPublicKey.Text := KeyToStr(e, n);
  FPrivateKey.Text := KeyToStr(d, n);

  FGIntDestroy(phi);
  FGIntDestroy(nilgint);
  FGIntDestroy(p);
  FGIntDestroy(q);
  FGIntDestroy(dp);
  FGIntDestroy(dq);
  FGIntDestroy(e);
  FGIntDestroy(d);
  FGIntDestroy(n);
end;
{$ENDIF}

procedure TXmlIni.Lock;
begin
  EnterCriticalSection(FLockCriSec);
end;

procedure TXmlIni.Unlock;
begin
  LeaveCriticalSection(FLockCriSec);
end;

function TXmlIni.SectionExists(const Section: WideString): Boolean;
begin
  Result := Exists(Section);
end;

function TXmlIni.ReadString(const Section, Ident, Default: WideString): WideString;
begin
  if Self.Section[Section].Exists(Ident) then
    Result := Self.Section[Section][Ident]
  else
    Result := Default;
end;

procedure TXmlIni.WriteString(const Section, Ident, Value: WideString);
begin
  Self.Section[Section][Ident] := Value;
end;

function TXmlIni.ReadInteger(const Section, Ident: WideString; Default: Longint): Longint;
begin
  if Self.Section[Section].Exists(Ident) then
    Result := Self.Section[Section].Key[Ident].AsInteger
  else
    Result := Default;
end;

procedure TXmlIni.WriteInteger(const Section, Ident: WideString; Value: Longint);
begin
  Self.Section[Section].Key[Ident].AsInteger := Value;
end;

function TXmlIni.ReadFloat(const Section, Ident: WideString; Default: Double): Double;
begin
  if Self.Section[Section].Exists(Ident) then
    Result := Self.Section[Section].Key[Ident].AsFloat
  else
    Result := Default;
end;

procedure TXmlIni.WriteFloat(const Section, Ident: WideString; Value: Double);
begin
  Self.Section[Section].Key[Ident].AsFloat := Value;
end;

function TXmlIni.ReadBool(const Section, Ident: WideString; Default: Boolean): Boolean;
begin
  if Self.Section[Section].Exists(Ident) then
    Result := Self.Section[Section].Key[Ident].AsBool
  else
    Result := Default;
end;

procedure TXmlIni.WriteBool(const Section, Ident: WideString; Value: Boolean);
begin
  Self.Section[Section].Key[Ident].AsBool := Value;
end;

function TXmlIni.ReadDateTime(const Section, Ident: WideString; Default: TDateTime): TDateTime;
begin
  if Self.Section[Section].Exists(Ident) then
    Result := Self.Section[Section].Key[Ident].AsDateTime
  else
    Result := Default;
end;

procedure TXmlIni.WriteDateTime(const Section, Ident: WideString; Value: TDateTime);
begin
  Self.Section[Section].Key[Ident].AsDateTime := Value;
end;

function TXmlIni.ReadDate(const Section, Ident: WideString; Default: TDateTime): TDateTime;
begin
  if Self.Section[Section].Exists(Ident) then
    Result := Self.Section[Section].Key[Ident].AsDate
  else
    Result := Default;
end;

procedure TXmlIni.WriteDate(const Section, Ident: WideString; Value: TDateTime);
begin
  Self.Section[Section].Key[Ident].AsDate := Value;
end;

function TXmlIni.ReadTime(const Section, Ident: WideString; Default: TDateTime): TDateTime;
begin
  if Self.Section[Section].Exists(Ident) then
    Result := Self.Section[Section].Key[Ident].AsTime
  else
    Result := Default;
end;

procedure TXmlIni.WriteTime(const Section, Ident: WideString; Value: TDateTime);
begin
  Self.Section[Section].Key[Ident].AsTime := Value;
end;

function TXmlIni.ReadBinaryStream(const Section, Ident: WideString; Value: TStream): Integer;
begin
  Result := Self.Section[Section].Key[Ident].AsStream.Size;
  if Result > 0 then
    Value.CopyFrom(Self.Section[Section].Key[Ident].AsStream, 0);
end;

procedure TXmlIni.WriteBinaryStream(const Section, Ident: WideString; Value: TStream);
begin
  Self.Section[Section].Key[Ident].AsStream := Value;
end;

procedure TXmlIni.ReadSection(const Section: WideString; Strings: TStrings);
begin
  ReadSectionValues(Section, Strings);
end;

procedure TXmlIni.ReadSections(Strings: TStrings);
var
  i: Integer;
begin
  Strings.BeginUpdate;
  try
    Strings.Clear;
    for i := 0 to Count - 1 do
      Strings.Add(SectionByIndex[i].Name);
  finally
    Strings.EndUpdate;
  end;
end;

procedure TXmlIni.ReadSectionValues(const Section: WideString; Strings: TStrings);
var
  i: Integer;
begin
  Strings.BeginUpdate;
  try
    Strings.Clear;
    if Exists(Section) then
      for i := 0 to Self.Section[Section].Count - 1 do
        Strings.Add(Self.Section[Section].KeyByIndex[i].Name + '=' + Self.Section[Section].KeyByIndex[i].Text);
  finally
    Strings.EndUpdate;
  end;
end;

procedure TXmlIni.EraseSection(const Section: WideString);
begin
  Self.Delete(Section);
end;

procedure TXmlIni.DeleteKey(const Section, Ident: WideString);
begin
  if SectionExists(Section) then
    Self.Section[Section].Key[Ident].Clear;
end;

procedure TXmlIni.UpdateFile;
begin
  SaveToFile;
end;

function TXmlIni.ValueExists(const Section, Ident: WideString): Boolean;
begin
  if SectionExists(Section) then
    Result := Self.Section[Section].Exists(Ident)
  else
    Result := False;
end;

procedure TXmlIni.InternalXmlLoadFromStream(Stream: TStream);
var
  Root: IXmlNode;
  Enc: TXmlIniEncodingType;
  XMLDom: IXMLDocument;
  i: Integer;
  j: Integer;
begin
  CoInitialize(nil);

  CoInitialize(nil);

  XMLDom := NewXMLDocument;
  try
    Stream.Position := 0;
    XMLDom.LoadFromStream(Stream, xetUnknown);

    Clear;

    FXmlEncoding := xietUnknow;

    for Enc := Low(XI_XML_EncodingStr) to High(XI_XML_EncodingStr) do
      if XMLDom.Encoding = XI_XML_EncodingStr[Enc] then
      begin
        FXmlEncoding := Enc;
        Break;
      end;

    Root := XMLDom.DocumentElement;

    if Root.NodeName <> XI_Root_NodeName then
      Exit;

    for i := 0 to Root.ChildNodes.Count - 1 do
      if Root.ChildNodes[i].NodeName = XI_Section_NodeName then
        with Add(Root.ChildNodes[i].Attributes[XI_Name_Attribute]) do
        begin
          for j := 0 to Root.ChildNodes[i].ChildNodes.Count - 1 do
            if Root.ChildNodes[i].ChildNodes[j].NodeName = XI_Key_NodeName then
              Item[Root.ChildNodes[i].ChildNodes[j].Attributes[XI_Name_Attribute]] :=
                Root.ChildNodes[i].ChildNodes[j].Text;
        end;
  finally
    XMLDom := nil;
  end;

  Collation;
end;

procedure TXmlIni.InternalXmlSaveToStream(Stream: TStream);
  function GetLocalCodePage: Integer;
  const
    CP_ACP = 0;
    LOCALE_IDEFAULTANSICODEPAGE = $00001004;
  var
    ResultCode: Integer;
    Buffer: array[0..6] of Char;
  begin
    GetLocaleInfo(GetThreadLocale, LOCALE_IDEFAULTANSICODEPAGE, Buffer, SizeOf(Buffer));
    Val(Buffer, Result, ResultCode);
    if ResultCode <> 0 then
      Result := CP_ACP;
  end;
  function GetLocalEncodingType: TXmlIniEncodingType;
  begin
    case GetLocalCodePage of
      950: Result := xietBIG5;
      936: Result := xietGB2312;
      932: Result := xietS_JIS;
      949: Result := xietEUC_KR;
    else
      Result := xietUTF_8;
    end;
  end;
var
  Root, SectionNode, KeyNode: IXmlNode;
  XMLDom: IXMLDocument;
  i: Integer;
  j: Integer;
begin
  Collation;

  if FXmlEncoding = xietUnknow then
    FXmlEncoding := GetLocalEncodingType;

  CoInitialize(nil);

  XMLDom := NewXMLDocument;
  try
    XMLDom.Options := XMLDom.Options + [doNodeAutoIndent];
    XMLDom.NodeIndentStr := '  ';

    Root := XMLDom.AddChild(XI_ROOT_NodeName);

    for i := 0 to Count - 1 do
    begin
      SectionNode := Root.AddChild(XI_Section_NodeName);
      SectionNode.Attributes[XI_Name_Attribute] := SectionByIndex[i].Name;
      for j := 0 to SectionByIndex[i].Count - 1 do
      begin
        KeyNode := SectionNode.AddChild(XI_Key_NodeName);
        KeyNode.Attributes[XI_Name_Attribute] := SectionByIndex[i].KeyByIndex[j].Name;
        KeyNode.Text := SectionByIndex[i].KeyByIndex[j].Text;
      end;
    end;

    XMLDom.StandAlone := XI_XML_StandAlone;
    XMLDom.Version := XI_XML_Version;
    XMLDom.Encoding := XI_XML_EncodingStr[FXmlEncoding];

    Stream.Position := 0;
    if Stream is TStringStream then
      TStringStream(Stream).WriteString(XMLDom.XML.Text)
    else
      XMLDom.SaveToStream(Stream);
  finally
    XMLDom := nil
  end;
end;

procedure TXmlIni.InternalLoadFromFile(AFileName: WideString);
var
  MS: TMemoryStream;
  FileLockMutex: TMutex;
begin
  Clear;

  MS := TMemoryStream.Create;
  try
    FileLockMutex := TMutex.Create(
      StringReplace(XI_LockMutexName + AFileName, '\', '_', [rfReplaceAll, rfIgnoreCase]));
    try
      if FileLockMutex.Get then
      begin
        try
          MS.LoadFromFile(AFileName);
        finally
          FileLockMutex.Release;
        end;
      end;
    finally
      FileLockMutex.Free;
    end;

    LoadFromStream(MS);
  finally
    MS.Free;
  end;
end;

procedure TXmlIni.InternalSaveToFile(AFileName: WideString);
var
  MS: TMemoryStream;
  FileLockMutex: TMutex;
begin
  MS := TMemoryStream.Create;
  try
    SaveToStream(MS);

    FileLockMutex := TMutex.Create(
      StringReplace(XI_LockMutexName + AFileName, '\', '_', [rfReplaceAll, rfIgnoreCase]));
    try
      if FileLockMutex.Get then
      begin
        try
          MS.SaveToFile(AFileName);
        finally
          FileLockMutex.Release;
        end;
      end;
    finally
      FileLockMutex.Free;
    end;
  finally
    MS.Free;
  end;
end;

function TXmlIni.InternalFind(SectionName: WideString): TXmlIniSection;
var
  i: Integer;
begin
  Result := nil;

  SectionName := Trim(SectionName);

  for i := Count - 1 downto 0 do
    if (SectionByIndex[i].Name = SectionName) or
      (not CaseSensitive and (UpperCase(SectionByIndex[i].Name) = UpperCase(SectionName))) then
    begin
      Result := SectionByIndex[i];
      Exit;
    end;
end;

function TXmlIni.InternalAdd(SectionName: WideString): TXmlIniSection;
begin
  Result := TXmlIniSection.Create(SectionName);
  Result.CaseSensitive := FCaseSensitive;
  FList.Add(Result);
end;

{$IFDEF RSA_ENCRYPT_ENABLED}
procedure TXmlIni.ExtractKey(KeyStr: string; var Key1, Key2: TFGInt);
var
  OutMS: TMemoryStream;
  InSS: TStringStream;
  i: Integer;
  KSize: ShortInt;
begin
  InSS := TStringStream.Create(KeyStr);
  try
    OutMS := TMemoryStream.Create;
    try
      MimeDecodeStream(InSS, OutMS);

      OutMS.Position := 0;

      OutMS.Read(KSize, SizeOf(KSize));
      if KSize >= 0 then
        Key1.Sign := Positive
      else
        Key1.Sign := Negative;
      KSize := Abs(KSize);
      SetLength(Key1.Number, KSize+1);
      Key1.Number[0] := KSize;
      for i := 1 to Length(Key1.Number)-1 do
        OutMS.Read(Key1.Number[i], SizeOf(LongWord));

      OutMS.Read(KSize, SizeOf(KSize));
      if KSize >= 0 then
        Key2.Sign := Positive
      else
        Key2.Sign := Negative;
      KSize := Abs(KSize);
      SetLength(Key2.Number, KSize+1);
      Key2.Number[0] := KSize;
      for i := 1 to Length(Key2.Number)-1 do
        OutMS.Read(Key2.Number[i], SizeOf(LongWord));
    finally
      OutMS.Free;
    end;
  finally
    InSS.Free;
  end;
end;
{$ENDIF}

function TXmlIni.GetCount: Integer;
begin
  Result := FList.Count;
end;

function TXmlIni.GetSectionByIndex(Index: Integer): TXmlIniSection;
begin
  if (Index < 0) or (Index >= FList.Count) then
    Result := GetSection('')
  else
    Result := TXmlIniSection(FList[Index]);
end;

function TXmlIni.GetSection(Index: WideString): TXmlIniSection;
begin
  Result := Add(Index);
end;

procedure TXmlIni.SetFileName(const Value: WideString);
begin
  if FFileName <> Value then
  begin
    if FAutoSave then
      SaveToFile;
    FFileName := Value;
    LoadFromFile;
  end;
end;

procedure TXmlIni.SetAutoSave(const Value: Boolean);
begin
  FAutoSave := Value;
  if FAutoSave then
    SaveToFile;
end;

procedure TXmlIni.SetCaseSensitive(const Value: Boolean);
var
  i: Integer;
begin
  if FCaseSensitive <> Value then
  begin
    FCaseSensitive := Value;

    for i := 0 to Count - 1 do
      SectionByIndex[i].CaseSensitive := FCaseSensitive;
  end;
end;

procedure TXmlIni.SetCompressed(const Value: Boolean);
var
  i: Integer;
begin
  if FCompressed <> Value then
  begin
    FCompressed := Value;

    if FAutoSave then
      SaveToFile;

    for i := 0 to Count - 1 do
      SectionByIndex[i].Compressed := FCompressed;
  end;
end;

procedure TXmlIni.SetEncryptMode(const Value: TXmlIniEncryptMode);
begin
  if FEncryptMode <> Value then
    FEncryptMode := Value;
end;

procedure TXmlIni.SetPassword(const Value: WideString);
begin
  if FPassword <> Trim(Value) then
    FPassword := Trim(Value);
end;

{$IFDEF RSA_ENCRYPT_ENABLED}
function TXmlIni.GetPrivateKey: TStrings;
begin
  Result := FPrivateKey;
end;

procedure TXmlIni.SetPrivateKey(const Value: TStrings);
begin
  FPrivateKey.Assign(Value);
end;

function TXmlIni.GetPublicKey: TStrings;
begin
  Result := FPublicKey;
end;

procedure TXmlIni.SetPublicKey(const Value: TStrings);
begin
  FPublicKey.Assign(Value);
end;
{$ENDIF}

procedure TXmlIni.SetRaiseException(const Value: Boolean);
var
  i: Integer;
begin
  if FRaiseException <> Value then
  begin
    FRaiseException := Value;
    for i := 0 to Count - 1 do
      SectionByIndex[i].RaiseException := FRaiseException;
  end;
end;

{ TXmlIniSection }

constructor TXmlIniSection.Create(AName: WideString = ''; AOwner: TXmlIni = nil);
begin
  FName := Trim(AName);
  FOwner := AOwner;
  FList := TList.Create;

  FXmlEncoding := xietUTF_16;
  FCaseSensitive := False;
  FCompressed := False;

  if AOwner <> nil then
    FRaiseException := FOwner.RaiseException
  else
    FRaiseException := False;
end;

destructor TXmlIniSection.Destroy;
begin
  Clear;
  FList.Free;

  inherited;
end;

function TXmlIniSection.Equals(Dest: TXmlIniSection): Boolean;
var
  i: Integer;
begin
  Collation;
  Dest.Collation;
  Result := False;
  for i := 0 to Count - 1 do
  begin
    if not Dest.Exists(KeyByIndex[i].Name) then
      Exit;
    if Dest[KeyByIndex[i].Name] <> KeyByIndex[i].Text then
      Exit;
  end;
  Result := True;
end;

procedure TXmlIniSection.CopyFrom(Source: TXmlIniSection);
begin
  Clear;
  Append(Source);
end;

procedure TXmlIniSection.Append(Source: TXmlIniSection);
var
  i: Integer;
begin
  for i := 0 to Source.Count - 1 do
    SetItem(Source.KeyByIndex[i].Name, Source.KeyByIndex[i].Text);
end;

function TXmlIniSection.Equals(Dest: TStrings): Boolean;
var
  i: Integer;
begin
  Result := False;
  for i := 0 to Count - 1 do
  begin
    if Dest.IndexOfName(KeyByIndex[i].Name) < 0 then
      Exit;
    if Dest.Values[KeyByIndex[i].Name] <> KeyByIndex[i].Text then
      Exit;
  end;
  Result := True;
end;

procedure TXmlIniSection.CopyFrom(Source: TStrings);
begin
  Clear;
  Append(Source);
end;

procedure TXmlIniSection.Append(Source: TStrings);
var
  i: Integer;
begin
  for i := 0 to Source.Count - 1 do
    SetItem(Source.Names[i], Source.Values[Source.Names[i]]);
end;

function TXmlIniSection.Exists(KeyName: WideString): Boolean;
begin
  Result := InternalFind(KeyName) <> nil;
end;

procedure TXmlIniSection.Clear;
var
  i: Integer;
begin
  FXmlEncoding := xietUTF_16;

  for i := Count - 1 downto 0 do
  begin
    KeyByIndex[i].Free;
    FList.Delete(i);
  end;
end;

procedure TXmlIniSection.Collation;
var
  i: Integer;
begin
  for i := Count - 1 downto 0 do
    if KeyByIndex[i].Text = '' then
    begin
      KeyByIndex[i].Free;
      FList.Delete(i);
    end;
end;

procedure TXmlIniSection.LoadFromStream(Stream: TStream; SectionName: WideString = '');
var
  XmlIni: TXmlIni;
begin
  XmlIni := TXmlIni.Create;
  try
    XmlIni.LoadFromStream(Stream);

    if (SectionName = '') and not XmlIni.Exists('') then
    begin
      if XmlIni.Exists(FName) then
        SectionName := FName
      else if XmlIni.Count > 0 then
        SectionName := XmlIni.SectionByIndex[0].Name;
    end;

    CopyFrom(XmlIni[SectionName]);
    FName := XmlIni[SectionName].Name;

    FXmlEncoding := XmlIni.XmlEncoding;
    FCompressed := XmlIni.Compressed;
  finally
    XmlIni.Free;
  end;
end;

procedure TXmlIniSection.SaveToStream(Stream: TStream; SectionName: WideString = '');
var
  XmlIni: TXmlIni;
begin
  if SectionName = '' then
    SectionName := FName;

  XmlIni := TXmlIni.Create;
  try
    XmlIni[SectionName].CopyFrom(Self);
    XmlIni.XmlEncoding := FXmlEncoding;
    XmlIni.Compressed := FCompressed;
    XmlIni.SaveToStream(Stream);
  finally
    XmlIni.Free;
  end;
end;

procedure TXmlIniSection.LoadFromString(Value: WideString; SectionName: WideString = '');
var
  SS: TStringStream;
begin
  if TrimLeft(Value) = '' then
    Clear
  else if TrimLeft(Value)[1] = '<' then
  begin
    SS := TStringStream.Create(Value);
    try
      LoadFromStream(SS, SectionName);
    finally
      SS.Free;
    end;
  end
  else
    LoadFromBase64String(Value, SectionName);
end;

function TXmlIniSection.SaveToString(SectionName: WideString = ''): WideString;
var
  SS: TStringStream;
begin
  if not FCompressed then
  begin
    SS := TStringStream.Create('');
    try
      SaveToStream(SS, SectionName);
      Result := SS.DataString;
    finally
      SS.Free;
    end;
  end
  else
    Result := SaveToBase64String;
end;

procedure TXmlIniSection.LoadFromBase64String(Value: string; SectionName: WideString = '');
var
  SS: TStringStream;
  MS: TMemoryStream;
begin
  if (TrimLeft(Value) = '') or (TrimLeft(Value)[1] = '<') then
    LoadFromString(Value)
  else
  begin
    SS := TStringStream.Create(Trim(Value));
    try
      MS := TMemoryStream.Create;
      try
        SS.Position := 0;
        MimeDecodeStream(SS, MS);
        LoadFromStream(MS, SectionName);
      finally
        MS.Free;
      end;
    finally
      SS.Free;
    end;
  end;
end;

function TXmlIniSection.SaveToBase64String(SectionName: WideString = ''): string;
var
  SS: TStringStream;
  MS: TMemoryStream;
begin
  MS := TMemoryStream.Create;
  try
    SaveToStream(MS, SectionName);

    SS := TStringStream.Create('');
    try
      MS.Position := 0;
      MimeEncodeStream(MS, SS);
      Result := SS.DataString;
    finally
      SS.Free;
    end;
  finally
    MS.Free;
  end;
end;

procedure TXmlIniSection.LoadFromFile(FileName: WideString; SectionName: WideString = '');
var
  XmlIni: TXmlIni;
begin
  XmlIni := TXmlIni.Create;
  try
    XmlIni.LoadFromFile(FileName);

    if (SectionName = '') and not XmlIni.Exists('') then
    begin
      if XmlIni.Exists(FName) then
        SectionName := FName
      else if XmlIni.Count > 0 then
        SectionName := XmlIni.SectionByIndex[0].Name;
    end;

    CopyFrom(XmlIni[SectionName]);
    FName := XmlIni[SectionName].Name;

    FXmlEncoding := XmlIni.XmlEncoding;
    FCompressed := XmlIni.Compressed;
  finally
    XmlIni.Free;
  end;
end;

procedure TXmlIniSection.SaveToFile(FileName: WideString; SectionName: WideString = '');
var
  XmlIni: TXmlIni;
begin
  if SectionName = '' then
    SectionName := FName;

  XmlIni := TXmlIni.Create;
  try
    XmlIni[SectionName].CopyFrom(Self);
    XmlIni.XmlEncoding := FXmlEncoding;
    XmlIni.Compressed := FCompressed;
    XmlIni.SaveToFile(FileName);
  finally
    XmlIni.Free;
  end;
end;

procedure TXmlIniSection.LoadFromRegistry(RootKey: HKEY; Path: WideString; SectionName: WideString = '');
var
  Reg: TRegistry;
  RegDataInfo: TRegDataInfo;
  Names: TStringList;
  i: Integer;
  Buf: PChar;
begin
  Clear;

  if (Path = '') or (Path[Length(Path)] <> '\') then
    Path := Path + '\';

  Reg := TRegistry.Create;
  try
    Reg.RootKey := RootKey;

    if SectionName <> '' then
      Path := Path + SectionName;

    Names := TStringList.Create;
    try
      if Reg.OpenKey(Path, False) then
      begin
        Reg.GetValueNames(Names);

        if Names.Count = 0 then
        begin
          Reg.CloseKey;
          if Reg.OpenKey(Path, False) then
            Reg.GetValueNames(Names);
        end;

        for i := 0 to Names.Count - 1 do
        begin
          Reg.GetDataInfo(Names[i], RegDataInfo);

          case RegDataInfo.RegData of
            rdString, rdExpandString:
              Key[Names[i]].Text := Reg.ReadString(Names[i]);
            rdInteger:
              Key[Names[i]].AsInteger := Reg.ReadInteger(Names[i]);
            rdUnknown, rdBinary:
              begin
                Buf := AllocMem(RegDataInfo.DataSize);
                try
                  Reg.ReadBinaryData(Names[i], Buf^, RegDataInfo.DataSize);
                  Key[Names[i]].FStream.Write(Buf^, RegDataInfo.DataSize);
                finally
                  FreeMem(Buf);
                end;
              end;
          end;
        end;

        Reg.CloseKey;
      end;

      Collation;
    finally
      Names.Free;
    end;
  finally
    Reg.Free;
  end;
end;

procedure TXmlIniSection.SaveToRegistry(RootKey: HKEY; Path: WideString; SectionName: WideString = '';
  UseOldDataType: Boolean = False);
var
  Reg: TRegistry;
  RegDataInfo: TRegDataInfo;
  Names: TStringList;
  i: Integer;
begin
  Collation;

  if (Path = '') or (Path[Length(Path)] <> '\') then
    Path := Path + '\';

  Reg := TRegistry.Create;
  try
    Reg.RootKey := RootKey;

    if SectionName <> '' then
      Path := Path + SectionName;

    Names := TStringList.Create;
    try
      if Reg.OpenKey(Path, True) then
      begin
        Reg.GetValueNames(Names);

        for i := 0 to Names.Count - 1 do
          if not UseOldDataType or not Exists(Names[i]) then
            Reg.DeleteValue(Names[i]);

        for i := 0 to Count - 1 do
        begin
          if UseOldDataType and Reg.ValueExists(KeyByIndex[i].Name) then
          begin
            Reg.GetDataInfo(KeyByIndex[i].Name, RegDataInfo);

            case RegDataInfo.RegData of
              rdString, rdExpandString:
                Reg.WriteString(KeyByIndex[i].Name, KeyByIndex[i].Text);
              rdInteger:
                Reg.WriteInteger(KeyByIndex[i].Name, KeyByIndex[i].AsInteger);
              rdUnknown, rdBinary:
                begin
                  KeyByIndex[i].TextToBinary;
                  Reg.WriteBinaryData(KeyByIndex[i].Name, KeyByIndex[i].FStream.Memory^, KeyByIndex[i].FStream.Size);
                end;
            else
            end;
          end
          else
          begin
            case KeyByIndex[i].DataType of
              xidtInteger:
                Reg.WriteInteger(KeyByIndex[i].Name, KeyByIndex[i].AsInteger);
              xidtText, xidtDateTime, xidtBool:
                Reg.WriteString(KeyByIndex[i].Name, KeyByIndex[i].Text);
              xidtBinary:
                begin
                  KeyByIndex[i].TextToBinary;
                  Reg.WriteBinaryData(KeyByIndex[i].Name, KeyByIndex[i].FStream.Memory^, KeyByIndex[i].FStream.Size);
                end;
            end;
          end;
        end;

        Reg.CloseKey;
      end;
    finally
      Names.Free;
    end;
  finally
    Reg.Free;
  end;
end;

procedure TXmlIniSection.LoadFromIni(FileName, SectionName: WideString);
var
  XmlIni: TXmlIni;
begin
  XmlIni := TXmlIni.Create;
  try
    XmlIni.LoadFromIni(FileName);

    if (SectionName = '') and not XmlIni.Exists('') then
    begin
      if XmlIni.Exists(FName) then
        SectionName := FName
      else if XmlIni.Count > 0 then
        SectionName := XmlIni.SectionByIndex[0].Name;
    end;

    CopyFrom(XmlIni[SectionName]);
    FName := XmlIni[SectionName].Name;

    FXmlEncoding := XmlIni.XmlEncoding;
    FCompressed := XmlIni.Compressed;
  finally
    XmlIni.Free;
  end;
end;

procedure TXmlIniSection.SaveToIni(FileName, SectionName: WideString);
var
  XmlIni: TXmlIni;
begin
  if SectionName = '' then
    SectionName := FName;

  XmlIni := TXmlIni.Create;
  try
    XmlIni[SectionName].CopyFrom(Self);
    XmlIni.XmlEncoding := FXmlEncoding;
    XmlIni.Compressed := FCompressed;
    XmlIni.SaveToFile(FileName);
  finally
    XmlIni.Free;
  end;
end;

function TXmlIniSection.InternalAdd(KeyName: WideString): TXmlIniKey;
begin
  Result := TXmlIniKey.Create(KeyName);
  FList.Add(Result);
end;

function TXmlIniSection.InternalFind(KeyName: WideString): TXmlIniKey;
var
  i: Integer;
begin
  Result := nil;

  KeyName := Trim(KeyName);

  for i := 0 to Count - 1 do
    if (KeyByIndex[i].Name = KeyName) or
      (not FCaseSensitive and (UpperCase(KeyByIndex[i].Name) = UpperCase(KeyName))) then
    begin
      Result := KeyByIndex[i];
      Break;
    end;
end;

procedure TXmlIniSection.SetName(const Value: WideString);
begin
  try
    if (FOwner <> nil) and not FOwner.Exists(Value) then
      raise Exception.Create('Section "' + Value + '" already exists.');
    FName := Value;
  except
    if FRaiseException then
      raise;
  end;
end;

function TXmlIniSection.GetKeyByIndex(Index: Integer): TXmlIniKey;
begin
  if (Index < 0) or (Index >= FList.Count) then
    Result := GetKey('')
  else
    Result := TXmlIniKey(FList[Index]);
end;

function TXmlIniSection.GetKey(Index: WideString): TXmlIniKey;
var
  Key: TXmlIniKey;
begin
  Key := InternalFind(Index);
  if Key = nil then
    Key := InternalAdd(Index);
  Result := Key;
end;

function TXmlIniSection.GetItem(Index: WideString): WideString;
begin
  Result := Key[Index].Text;
end;

procedure TXmlIniSection.SetItem(Index: WideString; const Value: WideString);
begin
  Key[Index].Text := Value;
end;

function TXmlIniSection.GetCount: Integer;
begin
  Result := FList.Count;
end;

procedure TXmlIniSection.SetRaiseException(const Value: Boolean);
var
  i: Integer;
begin
  if FRaiseException <> Value then
  begin
    FRaiseException := Value;
    for i := 0 to Count - 1 do
      KeyByIndex[i].RaiseException := FRaiseException;
  end;
end;

{ TXmlIniKey }

constructor TXmlIniKey.Create(AName: WideString = ''; AText: WideString = ''; AOwner: TXmlIniSection = nil);
begin
  FOwner := AOwner;
  FName := Trim(AName);
  FStream := TXmlIniEventMemoryStream.Create(Self);
  FDataType := xidtText;
  if AOwner <> nil then
    FRaiseException := FOwner.RaiseException
  else
    FRaiseException := False;
  Text := AText;
end;

destructor TXmlIniKey.Destroy;
begin
  FStream.Free;

  inherited;
end;

procedure TXmlIniKey.Clear;
begin
  FText := '';
  FStream.Clear;
  FDataType := xidtText;
end;

procedure TXmlIniKey.ValueLoadFromFile(FileName: WideString; Binary: Boolean = True);
var
  SS: TStringStream;
  MS: TMemoryStream;
begin
  try
    Clear;

    FileName := Trim(FileName);

    if (FileName <> '') and (ExtractFilePath(FileName) = '') then
      FileName := ExtractFilePath(ParamStr(0)) + FileName;

    if (FileName = '') or not FileExists(FileName) then
      if FRaiseException then
        raise Exception.Create('File "' + FileName + '" not exists.')
      else
        Exit;

    if Binary then
      FStream.LoadFromFile(FileName)
    else
    begin
      SS := TStringStream.Create('');
      try
        MS := TMemoryStream.Create;
        try
          MS.LoadFromFile(FileName);
          SS.CopyFrom(MS, 0);
        finally
          MS.Free;
        end;
        Text := SS.DataString;
      finally
        SS.Free;
      end;
    end;
  except
    if FRaiseException then
      raise;
  end;
end;

procedure TXmlIniKey.ValueSaveToFile(FileName: WideString);
var
  SS: TStringStream;
  MS: TMemoryStream;
begin
  FileName := Trim(FileName);

  if (FileName <> '') and (ExtractFilePath(FileName) = '') then
    FileName := ExtractFilePath(ParamStr(0)) + FileName;

  if FileName = '' then
    if FRaiseException then
      raise Exception.Create('File name is required.')
    else
      Exit;

  try
    if DataType = xidtBinary then
    begin
      if FStream.Size > 0 then
        FStream.SaveToFile(FileName);
    end
    else
    begin
      if FText <> '' then
      begin
        SS := TStringStream.Create(FText);
        try
          MS := TMemoryStream.Create;
          try
            MS.CopyFrom(SS, 0);
            MS.SaveToFile(FileName);
          finally
            MS.Free;
          end;
        finally
          SS.Free;
        end;
      end;
    end;
  except
    if FRaiseException then
      raise;
  end;
end;

procedure TXmlIniKey.SetName(const Value: WideString);
begin
  try
    if (FOwner <> nil) and not FOwner.Exists(Value) then
      raise Exception.Create('Key "' + Value + '" already exists.');
    FName := Value;
  except
    if FRaiseException then
      raise;
  end;
end;

procedure TXmlIniKey.BinaryToText;
var
  SS: TStringStream;
begin
  try
    if DataType = xidtBinary then
    begin
      SS := TStringStream.Create('');
      try
        FStream.Position := 0;
        MimeEncodeStream(FStream, SS);
        FText := XI_Base64Header + SS.DataString;
      finally
        SS.Free;
      end;
    end
    else
    begin
      SS := TStringStream.Create('');
      try
        SS.CopyFrom(FStream, 0);
        FText := SS.DataString;
      finally
        SS.Free;
      end;
    end;
  except
    if FRaiseException then
      raise;
  end;
end;

procedure TXmlIniKey.TextToBinary;
var
  SS: TStringStream;
  Base64Text: WideString;
  OldDataType: TXmlIniDataType;
begin
  try
    if Copy(FText, 1, Length(XI_Base64Header)) <> XI_Base64Header then
    begin
      SS := TStringStream.Create(FText);
      try
        OldDataType := DataType;
        FStream.CopyFrom(SS, 0);
        FDataType := OldDataType;
      finally
        SS.Free;
      end;
    end
    else
    begin
      Base64Text := Trim(FText);

      Delete(Base64Text, 1, 8);

      SS := TStringStream.Create(Base64Text);
      try
        MimeDecodeStream(SS, FStream);
      finally
        SS.Free;
      end;
    end;
  except
    if FRaiseException then
      raise;
  end;
end;

function TXmlIniKey.GetText: WideString;
begin
  if (FText = '') and (FStream.Size > 0) then
    BinaryToText;

  Result := FText;
end;

procedure TXmlIniKey.SetText(const Value: WideString);
begin
  if FText <> Value then
  begin
    FStream.Clear;

    FText := Value;
    FDataType := xidtUnknow;
  end;
end;

function TXmlIniKey.GetAsStream: TStream;
begin
  Result := FStream;
end;

procedure TXmlIniKey.SetAsStream(const Value: TStream);
begin
  FText := '';

  if (Value <> nil) and (Value.Size > 0) then
  begin
    Value.Position := 0;
    FStream.CopyFrom(Value, 0);
    FDataType := xidtBinary;
  end
  else
  begin
    FStream.Clear;
    FDataType := xidtText;
  end;
end;

function TXmlIniKey.GetAsInteger: Integer;
begin
  Result := StrToIntDef(Text, 0);
end;

procedure TXmlIniKey.SetAsInteger(const Value: Integer);
begin
  Text := IntToStr(Value);
  FDataType := xidtInteger;
end;

function TXmlIniKey.GetAsFloat: Double;
begin
  Result := StrToFloatDef(Text, 0);
end;

procedure TXmlIniKey.SetAsFloat(const Value: Double);
begin
  Text := FloatToStr(Value);
  FDataType := xidtFloat;
end;

function TXmlIniKey.GetAsBool: Boolean;
var
  TestStr: WideString;
  i: Integer;
begin
  TestStr := UpperCase(Trim(Text));

  Result := False;

  for i := Low(XI_True_Text) to High(XI_True_Text) do
    if UpperCase(XI_True_Text[i]) = TestStr then
    begin
      Result := True;
      Break;
    end;
end;

procedure TXmlIniKey.SetAsBool(const Value: Boolean);
begin
  if Value then
    Text := XI_True_Text[0]
  else
    Text := XI_False_Text[0];
  FDataType := xidtBool;
end;

function TXmlIniKey.GetAsDateTime: TDateTime;
begin
  Result := ExStrToDateTime(Text);
end;

procedure TXmlIniKey.SetAsDateTime(const Value: TDateTime);
begin
  Text := ExDateTimeToStr(Value);
  FDataType := xidtDateTime;
end;

function TXmlIniKey.GetAsDate: TDateTime;
begin
  Result := ExStrToDate(Text);
end;

procedure TXmlIniKey.SetAsDate(const Value: TDateTime);
begin
  Text := ExDateToStr(Value);
  FDataType := xidtDate;
end;

function TXmlIniKey.GetAsTime: TDateTime;
begin
  Result := ExStrToTime(Text);
end;

procedure TXmlIniKey.SetAsTime(const Value: TDateTime);
begin
  Text := ExTimeToStr(Value);
  FDataType := xidtTime;
end;

function TXmlIniKey.GetDataType: TXmlIniDataType;
  function InArray(Str: WideString; ArrayStr: array of WideString): Boolean;
  var
    i: Integer;
  begin
    Result := False;

    Str := UpperCase(Str);

    for i := Low(ArrayStr) to High(ArrayStr) do
      if Str = UpperCase(ArrayStr[i]) then
      begin
        Result := True;
        Break;
      end;
  end;
  function IsAllNumber(Str: WideString): Boolean;
  var
    i: Integer;
  begin
    Result := True;
    Str := Trim(Str);
    if Str <> '' then
      for i := 1 to Length(Str) do
        if not ((Str[i] >= '0') and (Str[i] < '9')) then
        begin
          Result := False;
          Break;
        end;
  end;
  function IsFloat(Str: WideString): Boolean;
  begin
    Str := Trim(Str);
    Result := Str = FloatToStr(StrToFloatDef(Str, 0));
  end;
  function IsDateTime(Str: WideString): Boolean;
  begin
    Str := Trim(Str);
    Result := Str = ExDateTimeToStr(ExStrToDateTime(Str));
  end;
  function IsDate(Str: WideString): Boolean;
  begin
    Str := Trim(Str);
    Result := Str = ExDateToStr(ExStrToDate(Str));
  end;
  function IsTime(Str: WideString): Boolean;
  begin
    Str := Trim(Str);
    Result := Str = ExDateToStr(ExStrToTime(Str));
  end;
begin
  if FDataType = xidtUnknow then
  begin
    if (FText <> '') and (Copy(FText, 1, Length(XI_Base64Header)) = XI_Base64Header) then
      FDataType := xidtBinary
    else if IsAllNumber(FText) then
      FDataType := xidtInteger
    else if InArray(FText, XI_True_Text) or InArray(FText, XI_False_Text) then
      FDataType := xidtBool
    else if IsFloat(FText) then
      FDataType := xidtFloat
    else if IsDateTime(FText) then
      FDataType := xidtDateTime
    else if IsDate(FText) then
      FDataType := xidtDate
    else if IsTime(FText) then
      FDataType := xidtTime
    else
      FDataType := xidtText;
  end;
  Result := FDataType;
end;

procedure TXmlIniKey.SetDataType(const Value: TXmlIniDataType);
var
  SS: TStringStream;
begin
  if DataType <> Value then
  begin
    if Value = xidtBinary then
    begin
      FStream.Clear;

      try
        SS := TStringStream.Create(FText);
        try
          FStream.CopyFrom(SS, 0);
        finally
          SS.Free;
        end;
      except
        if FRaiseException then
          raise;
      end;
    end
    else if DataType = xidtBinary then
    begin
      try
        if FStream.Size = 0 then
          FText := ''
        else
        begin
          SS := TStringStream.Create('');
          try
            SS.CopyFrom(FStream, 0);
            FText := SS.DataString;
          finally
            SS.Free;
          end;
        end;
        FStream.Clear;
      except
        if FRaiseException then
          raise;
      end;
    end;

    FDataType := Value;

    case FDataType of
      xidtText:
        Text := Text;
      xidtInteger:
        AsInteger := AsInteger;
      xidtFloat:
        AsFloat := AsFloat;
      xidtBool:
        AsBool := AsBool;
      xidtDateTime:
        AsDateTime := AsDateTime;
      xidtDate:
        AsDate := AsDate;
      xidtTime:
        AsTime := AsTime;
    end;
  end;
end;

{ TXmlIniEventMemoryStream }

constructor TXmlIniEventMemoryStream.Create(AOwner: TXmlIniKey);
begin
  FOwner := AOwner;

  inherited Create;
end;

function TXmlIniEventMemoryStream.GetSize: Int64;
begin
  if (inherited GetSize = 0) and (FOwner.FText <> '') then
    FOwner.TextToBinary;

  Result := inherited GetSize;
end;

procedure TXmlIniEventMemoryStream.SetSize(NewSize: Integer);
begin
  FOwner.FDataType := xidtBinary;
  FOwner.FText := '';

  inherited SetSize(NewSize);
end;

function TXmlIniEventMemoryStream.Read(var Buffer;
  Count: Integer): Longint;
begin
  if (inherited GetSize = 0) and (FOwner.FText <> '') then
    FOwner.TextToBinary;

  Result := inherited Read(Buffer, Count);
end;

function TXmlIniEventMemoryStream.Write(const Buffer;
  Count: Integer): Longint;
begin
  FOwner.FDataType := xidtBinary;
  FOwner.FText := '';

  Result := inherited Write(Buffer, Count);
end;

end.

